# coding=utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# noinspection PyUnresolvedReferences
from win32api import *

# Try and use XP features, so we get alpha-blending etc.
try:
    from winxpgui import *
except ImportError:
    # noinspection PyUnresolvedReferences
    from win32gui import *
# noinspection PyUnresolvedReferences
import win32con
import sys
import os
import struct
import time
# noinspection PyUnresolvedReferences
import win32com.client as com
# noinspection PyUnresolvedReferences
import win32event as w32e
# noinspection PyUnresolvedReferences
import mmapfile as mmf
# noinspection PyUnresolvedReferences
import win32api as win_api


def show_msg(title, msg):
    """
    windows pop弹窗，主要用在长时间且耗时的任务中，提示重要问题信息
    :param title: 弹窗标题
    :param msg: 弹窗信息
    """
    MainWindow(title, msg)


def socket_bind_recv(socket_fn, cmd_handler):
    """
    非bsd系统的进程间通信，接受消息，处理消息，使用windows全局共享内存实现，
    函数名称保持与bsd的接口名称一致
    :param socket_fn: 共享内存文件名称
    :param cmd_handler: cmd处理函数，callable类型
    """
    if not callable(cmd_handler):
        print('socket_bind_recv cmd_handler must callable!')

    while True:
        global_fn = 'Global\\{}'.format(socket_fn)
        event = w32e.CreateEvent(None, 0, 0, global_fn)
        event_mmap = mmf.mmapfile(None, socket_fn, 1024)
        w32e.WaitForSingleObject(event, -1)
        socket_cmd = event_mmap.read(1024).decode()
        # 把接收到的socket传递给外部对应的处理函数
        cmd_handler(socket_cmd)
        event_mmap.close()
        win_api.CloseHandle(event)


def socket_send_msg(socket_fn, msg):
    """
    非bsd系统的进程间通信，发送消息，使用windows全局共享内存实现，函数名称保持与bsd的接口名称一致
    :param socket_fn: : 共享内存名称
    :param msg: 字符串类型需要传递的数据，不需要encode，内部进行encode
    """
    global_fn = 'Global\\{}'.format(socket_fn)
    event = w32e.OpenEvent(w32e.EVENT_ALL_ACCESS, 0, global_fn)
    event_mmap = mmf.mmapfile(None, socket_fn, 1024)
    w32e.SetEvent(event)
    event_mmap.write(msg)
    event_mmap.close()
    win_api.CloseHandle(event)


def fold_free_size_mb(folder):
    """
    windows os下剩余磁盘空间获取
    :param folder: 目标目录
    :return: 返回float，单位mb
    """
    return drive_free_space(folder) / 1024 / 1024 / 1024


def drive_free_space(drive):
    # noinspection PyBroadException
    try:
        fso = com.Dispatch("Scripting.FileSystemObject")
        drv = fso.GetDrive(drive)
        return drv.FreeSpace
    except:
        return 0


def max_drive():
    space_array = dict()
    for i in range(65, 91):
        vol = chr(i) + '://'
        if os.path.isdir(vol):
            space_array[vol] = drive_free_space(vol)

    max_v = max(zip(space_array.values(), space_array.keys()))[1]
    if max_v.startswith('c'):
        return os.path.expanduser('~')
    return max_v


# noinspection PyClassHasNoInit
class PyNOTIFYICONDATA:
    _struct_format = (
        "I"  # DWORD cbSize;
        "I"  # HWND hWnd;
        "I"  # UINT uID;
        "I"  # UINT uFlags;
        "I"  # UINT uCallbackMessage;
        "I"  # HICON hIcon;
        "128s"  # TCHAR szTip[128];
        "I"  # DWORD dwState;
        "I"  # DWORD dwStateMask;
        "256s"  # TCHAR szInfo[256];
        "I"  # union {
        #    UINT  uTimeout;
        #    UINT  uVersion;
        # } DUMMYUNIONNAME;
        "64s"  # TCHAR szInfoTitle[64];
        "I"  # DWORD dwInfoFlags;
        #       GUID guidItem;
    )
    _struct = struct.Struct(_struct_format)

    hWnd = 0
    uID = 0
    uFlags = 0
    uCallbackMessage = 0
    hIcon = 0
    szTip = ''
    dwState = 0
    dwStateMask = 0
    szInfo = ''
    uTimeoutOrVersion = 0
    szInfoTitle = ''
    dwInfoFlags = 0

    def pack(self):
        return self._struct.pack(
            self._struct.size,
            self.hWnd,
            self.uID,
            self.uFlags,
            self.uCallbackMessage,
            self.hIcon,
            self.szTip,
            self.dwState,
            self.dwStateMask,
            self.szInfo,
            self.uTimeoutOrVersion,
            self.szInfoTitle,
            self.dwInfoFlags)

    def __setattr__(self, name, value):
        # avoid wrong field names
        if not hasattr(self, name):
            raise NameError(name)
        self.__dict__[name] = value


# noinspection PyUnresolvedReferences,PyUnusedLocal
class MainWindow:
    def __init__(self, title, msg):
        message_map = {
            win32con.WM_DESTROY: self.on_destroy,
        }
        # Register the Window class.
        wc = WNDCLASS()
        hinst = wc.hInstance = GetModuleHandle(None)
        wc.lpszClassName = "PythonTaskbarDemo"
        wc.lpfnWndProc = message_map  # could also specify a wndproc.
        class_atom = RegisterClass(wc)
        # Create the Window.
        style = win32con.WS_OVERLAPPED | win32con.WS_SYSMENU
        self.hwnd = CreateWindow(class_atom, "Taskbar Demo", style,
                                 0, 0, win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT,
                                 0, 0, hinst, None)
        UpdateWindow(self.hwnd)
        icon_path_name = os.path.abspath(os.path.join(sys.prefix, "pyc.ico"))
        icon_flags = win32con.LR_LOADFROMFILE | win32con.LR_DEFAULTSIZE
        # noinspection PyBroadException
        try:
            hicon = LoadImage(hinst, icon_path_name, win32con.IMAGE_ICON, 0, 0, icon_flags)
        except:
            hicon = LoadIcon(0, win32con.IDI_APPLICATION)
        flags = NIF_ICON | NIF_MESSAGE | NIF_TIP
        nid = (self.hwnd, 0, flags, win32con.WM_USER + 20, hicon, "Balloon  tooltip demo")
        Shell_NotifyIcon(NIM_ADD, nid)
        self.show_balloon(title, msg)
        time.sleep(20)
        DestroyWindow(self.hwnd)

    def show_balloon(self, title, msg):
        # For this message I can't use the win32gui structure because
        # it doesn't declare the new, required fields
        nid = PyNOTIFYICONDATA()
        nid.hWnd = self.hwnd
        nid.uFlags = NIF_INFO
        # type of balloon and text are random
        nid.dwInfoFlags = NIIF_INFO
        nid.szInfo = msg
        nid.szInfoTitle = title
        # Call the Windows function, not the wrapped one
        from ctypes import windll
        shell_notify_icon = windll.shell32.Shell_NotifyIconA
        shell_notify_icon(NIM_MODIFY, nid.pack())

    def on_destroy(self, hwnd, msg, wparam, lparam):
        nid = (self.hwnd, 0)
        Shell_NotifyIcon(NIM_DELETE, nid)
        # Terminate the app.
        PostQuitMessage(0)



from __future__ import absolute_import

from . import ABuNDAtr as atr
from . import ABuNDBoll as boll
from . import ABuNDMa as ma
from . import ABuNDMacd as macd
from . import ABuNDRsi as rsi

__all__ = [
    'atr',
    'boll',
    'ma',
    'macd',
    'rsi'
]



# -*- encoding:utf-8 -*-
"""
    包装选股worker进行，完善前后工作
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from .ABuPickStockWorker import AbuPickStockWorker
from ..CoreBu.ABuEnvProcess import add_process_env_sig
from ..MarketBu.ABuMarket import split_k_market
from ..TradeBu.ABuKLManager import AbuKLManager
from ..CoreBu.ABuFixes import ThreadPoolExecutor

__author__ = '阿布'
__weixin__ = 'abu_quant'


@add_process_env_sig
def do_pick_stock_work(choice_symbols, benchmark, capital, stock_pickers):
    """
    包装AbuPickStockWorker进行选股
    :param choice_symbols: 初始备选交易对象序列
    :param benchmark: 交易基准对象，AbuBenchmark实例对象
    :param capital: 资金类AbuCapital实例化对象
    :param stock_pickers: 选股因子序列
    :return:
    """
    kl_pd_manager = AbuKLManager(benchmark, capital)
    stock_pick = AbuPickStockWorker(capital, benchmark, kl_pd_manager, choice_symbols=choice_symbols,
                                    stock_pickers=stock_pickers)
    stock_pick.fit()
    return stock_pick.choice_symbols


@add_process_env_sig
def do_pick_stock_thread_work(choice_symbols, benchmark, capital, stock_pickers, n_thread):
    """包装AbuPickStockWorker启动线程进行选股"""
    result = []

    def when_thread_done(r):
        result.extend(r.result())

    with ThreadPoolExecutor(max_workers=n_thread) as pool:
        thread_symbols = split_k_market(n_thread, market_symbols=choice_symbols)
        for symbols in thread_symbols:
            future_result = pool.submit(do_pick_stock_work, symbols, benchmark, capital, stock_pickers)
            future_result.add_done_callback(when_thread_done)

    return result



# -*- encoding:utf-8 -*-
"""
    平台信息工具模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import platform
import sys
import struct
import logging


def is_32bit():
    """是否32位操作系统"""
    return struct.calcsize("P") * 8 < 64


def is_mac():
    """是否mac os操作系统"""
    return 'Darwin' in platform.system()


def is_windows():
    """是否Windows操作系统"""
    return 'Windows' in platform.system()


def is_linux():
    """是否Linux操作系统"""
    return 'Linux' in platform.system()


def platform_name():
    """操作系统名称字符串"""
    return platform.system()


def get_sys_info():
    """
        平台基本信息收集
        eg:
            [('python', '3.6.0.final.0'),
             ('python-bits', 64),
             ('OS', 'Darwin'),
             ('OS-release', '15.6.0'),
             ('machine', 'x86_64'),
             ('processor', 'i386'),
             ('byteorder', 'little'),
             ('LC_ALL', 'None'),
             ('LANG', 'zh_CN.UTF-8')]
    """
    sys_info = list()
    try:
        sys_name, node_name, release, version, machine, processor = platform.uname(
        )
        # noinspection PyStringFormat
        sys_info.extend([
            ("python", "%d.%d.%d.%s.%s" % sys.version_info[:]),
            ("python-bits", struct.calcsize("P") * 8),
            ("OS", "%s" % sys_name),
            ("OS-release", "%s" % release),
            ("machine", "%s" % machine),
            ("processor", "%s" % processor),
            ("byteorder", "%s" % sys.byteorder),
            ("LC_ALL", "%s" % os.environ.get('LC_ALL', "None")),
            ("LANG", "%s" % os.environ.get('LANG', "None")),
        ])
    except Exception as e:
        logging.exception(e)

    return sys_info


# noinspection PyDeprecation
def show_versions():
    """
        平台基本信息收集以及主要lib版本号信息
        eg.
            INSTALLED VERSIONS
            ------------------
            python: 3.6.0.final.0
            python-bits: 64
            OS: Darwin
            OS-release: 15.6.0
            machine: x86_64
            processor: i386
            byteorder: little
            LC_ALL: None
            LANG: zh_CN.UTF-8

            pandas: 0.19.2
            sklearn: 0.18.1
            numpy: 1.11.3
            scipy: 0.18.1
            statsmodels: 0.6.1
            notebook: 4.3.1
            tables: 3.3.0
            seaborn: 0.7.1
            matplotlib: 2.0.0
            requests: 2.12.4
            bs4: 4.5.3
            numba: 0.30.1
    """
    sys_info = get_sys_info()

    deps_mod = [
        # (MODULE_NAME, f(dep_mod) -> dep_mod version)
        ("pandas", lambda dep_mod: dep_mod.__version__),
        ("sklearn", lambda dep_mod: dep_mod.__version__),
        ("numpy", lambda dep_mod: dep_mod.version.version),
        ("scipy", lambda dep_mod: dep_mod.version.version),
        ("statsmodels", lambda dep_mod: dep_mod.__version__),
        ("notebook", lambda dep_mod: dep_mod.__version__),
        ("tables", lambda dep_mod: dep_mod.__version__),
        ("seaborn", lambda dep_mod: dep_mod.__version__),
        ("matplotlib", lambda dep_mod: dep_mod.__version__),
        ("requests", lambda dep_mod: dep_mod.__version__),
        ("bs4", lambda dep_mod: dep_mod.__version__),
        ("numba", lambda dep_mod: dep_mod.__version__)
    ]

    deps_info = list()
    for (modname, ver_f) in deps_mod:
        try:
            import imp
            try:
                mod = imp.load_module(modname, *imp.find_module(modname))
            except ImportError:
                import importlib
                mod = importlib.import_module(modname)
            ver = ver_f(mod)
            deps_info.append((modname, ver))
        except:
            deps_info.append((modname, None))
    print("\nINSTALLED VERSIONS")
    print("------------------")

    for k, stat in sys_info:
        print("%s: %s" % (k, stat))

    print("")
    for k, stat in deps_info:
        print("%s: %s" % (k, stat))


if __name__ == "__main__":
    show_versions()



# -*- encoding:utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__author__ = '小青蛙'
__weixin__ = 'abu_quant'

"""
s:static
d:dynamic
co: company
pb: price book ratio
pe: price earning ratio
ps: price sales ratio
asset : net_asset_value_per_share
oo:organization_ownership
cc:circulation_capital
mv:market_value
sc:short_covering
"""
columns_map = {
    "symbol": 'symbol',
    "market": 'market',
    "exchange": 'exchange',

    "industry": 'industry',
    "name": 'co_name',
    "公司网站:": 'co_site',
    "公司电话：": 'co_tel',
    "公司地址：": 'co_addr',
    "业务:": 'co_business',
    "简介:": 'co_intro',

    "市盈率(静)/(动)：": 'pe_s_d',
    "市净率(动)：": 'pb_d',
    "市净率MRQ：": 'pb_MRQ',
    "市销率(动)：": 'ps_d',
    "市销率：": 'ps',
    "市盈率(静)：": 'pe_s',
    "换手率：": 'turnover',

    "每股净资产：": 'asset',
    "流通股本：": 'cc',
    "振幅：": 'amplitude',
    "空头补回天数：": 'sc',
    "招股说明书：": 'prospectus',
    "总股本：": 'equity',
    "总市值：": 'mv',
    "机构持股：": 'oo',
    "港股股本：": 'hk_equity'
}



# -*- encoding:utf-8 -*-
"""
    相关系数，相似度可视化模块
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import itertools
import math

import matplotlib.pyplot as plt
import numpy as np

from ..MarketBu import ABuSymbolPd
from ..UtilBu import ABuScalerUtil

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import xrange
from ..CoreBu import ABuEnv

"""预备颜色序列集，超出序列数量应使用itertools.cycle循环绘制"""
K_PLT_MAP_STYLE = [
    'b', 'c', 'g', 'k', 'm', 'r', 'y', 'b--', 'c--', 'g--', 'k--']


def draw_show_close(sorted_ret, target_count, show_cnt):
    """
    通过多少个交易日参数target_count，计算出make_kl_df的参数n_folds，
    使用ABuScalerUtil.scaler_std将show_cnt个最相似的股票价格序列进行
    标准化在一个数量值范围内可视化
    :param sorted_ret: 可迭代序列，元素形如('usTSLA', 1.0), ('usSINA', 0.45565379371028253).....
    :param target_count: 需要请求多少个交易日数据，int
    :param show_cnt: 可视化top show_cnt相关个价格走势
    """
    if show_cnt is None and not isinstance(show_cnt, int):
        return
    # 规避sorted_ret长度不够的问题
    show_cnt = min(show_cnt, len(sorted_ret))
    if show_cnt <= 0:
        return

    # 循环K_PLT_MAP_STYLE颜色集的颜色，绘制各个金融时间序列
    for x, cs_color in zip(xrange(0, show_cnt), itertools.cycle(K_PLT_MAP_STYLE)):
        # 通过多少个交易日参数target_count，计算出要请求几年的数据n_folds
        n_folds = int(math.ceil(target_count / ABuEnv.g_market_trade_year))
        # sorted_ret[x] : ('usTSLA', 1.0) -> sorted_ret[x][0]: usTSLA
        # FIXME 暂时忽略一个bug如果请求可视化时使用的是start，end，方式那么这里可视化的时间段就不符合了，需要传递完整的信息
        df = ABuSymbolPd.make_kl_df(sorted_ret[x][0], n_folds=n_folds)
        # 支可视化close_array
        close_array = df['close']
        if target_count < len(close_array):
            # 再次确认时间范围
            close_array = close_array[:int(target_count)]

        cs_np = np.array(close_array, dtype=np.float)
        # 使用ABuScalerUtil.scaler_std将序列进行标准化在一个数量值范围内可视化
        plt.plot(ABuScalerUtil.scaler_std(cs_np), cs_color, label=sorted_ret[x][0])
    plt.legend(loc='best', bbox_to_anchor=(1.05, 1), borderaxespad=0.)
    plt.show()


def draw_show_close_array(sorted_ret, cs_array):
    """
    绘制参数中cs_array序列个金融时间序列
    :param sorted_ret: 可迭代序列，元素形如('usTSLA', 1.0), ('usSINA', 0.45565379371028253).....
    :param cs_array: 可迭代的价格序列
    """
    # 循环K_PLT_MAP_STYLE颜色集的颜色，绘制各个金融时间序列
    for x, (cs_np, cs_color) in enumerate(zip(cs_array, itertools.cycle(K_PLT_MAP_STYLE))):
        # sorted_ret[x] : ('usTSLA', 1.0) -> sorted_ret[x][0]: usTSLA
        plt.plot(cs_np, cs_color, label=sorted_ret[x][0], bbox_to_anchor=(1.05, 1), loc=3, borderaxespad=0.)

    plt.legend(loc='best')
    plt.show()



# -*- encoding:utf-8 -*-
"""
    量化波动程度模块
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from ..TLineBu.ABuTLine import AbuTLine
from ..CoreBu.ABuPdHelper import pd_rolling_std, pd_ewm_mean, pd_ewm_std, pd_resample
from ..UtilBu import ABuStatsUtil


def show_wave_return(kl_pd):
    """
        可视化收益的移动平均std和加权移动平均std
        注意会修改kl_pd，只做测试使用，内部未做copy处理，
        如不能改动，外部自copy操作，再传递进来
        :param kl_pd: 金融时间序列，pd.DataFrame对象
    """

    pre_close = kl_pd['close'].shift(1)
    # noinspection PyTypeChecker
    kl_pd['return'] = np.where(pre_close == 0, 0, np.log(kl_pd['close'] / pre_close))
    kl_pd['mov_std'] = pd_rolling_std(kl_pd['return'], window=20, center=False) * np.sqrt(20)
    kl_pd['std_ewm'] = pd_ewm_std(kl_pd['return'], span=20, min_periods=20, adjust=True) * np.sqrt(20)
    kl_pd[['close', 'mov_std', 'std_ewm', 'return']].plot(subplots=True, figsize=(16, 12), grid=True)
    plt.show()


def calc_wave_std(kl_pd, xd=21, ewm=True, show=True):
    """
    计算收益的移动平均std或者加权移动平均std技术线，使用
    AbuTLine封装技术线实体，不会修改kl_pd，返回AbuTLine对象
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param xd: 计算移动平均std或加权移动平均std使用的窗口参数，默认21
    :param ewm: 是否使用加权移动平均std计算
    :param show: 是否可视化，可视化使用AbuTLine.show接口
    :return: 返回AbuTLine对象
    """

    pre_close = kl_pd['close'].shift(1)
    # noinspection PyTypeChecker
    change = np.where(pre_close == 0, 0, np.log(kl_pd['close'] / pre_close))
    if ewm:
        roll_std = pd_ewm_std(change, span=xd, min_periods=1, adjust=True) * np.sqrt(xd)
    else:
        roll_std = pd_rolling_std(change, window=xd, min_periods=1, center=False) * np.sqrt(xd)

    # min_periods=1还是会有两个nan，填了
    roll_std = pd.Series(roll_std).fillna(method='bfill')
    # 主要目的就是通过roll_std构造AbuTLine对象line
    line = AbuTLine(roll_std, 'wave std')
    if show:
        line.show()

    return line


def calc_wave_abs(kl_pd, xd=21, show=True):
    """
    计算金融时间序列kl_pd在的绝对波动，通过参数xd对波动进行重采样
    在默认xd＝21情况下，变成了月震荡幅度，使用ABuStatsUtil.demean对
    月震荡幅度进行去均值操作后得到技术线demean_wave，AbuTLine包裹
    技术线返回
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param xd: 对波动进行重采样的周期，单位天，int
    :param show: 是否可视化
    :return: 返回AbuTLine对象
    """
    # 不考虑正负，只考虑波动，np.abs(kl_pd['p_change'])
    abs_pct_change = np.abs(kl_pd['p_change'])
    xd_resample = '%dD' % xd
    # 通过pd_resample重采样，使用how=sum, 即默认xd＝21情况下，变成了月震荡幅度
    change_ratio_sum = pd_resample(abs_pct_change, xd_resample, how='sum')
    """
        eg: change_ratio_sum
            2014-07-24    37.13
            2014-08-14    39.33
            2014-09-04    25.16
            2014-09-25    27.53
            2014-10-16    27.78
                          ...
            2016-04-14    25.17
            2016-05-05    42.07
            2016-05-26    18.93
            2016-06-16    33.25
            2016-07-07    10.79
    """
    # 使用ABuStatsUtil.demean进行去均值操作
    demean_wave = ABuStatsUtil.demean(change_ratio_sum)
    """
        eg: demean_wave
        2014-07-24    -1.6303
        2014-08-14     0.5697
        2014-09-04   -13.6003
        2014-09-25   -11.2303
                       ...
        2016-05-05     3.3097
        2016-05-26   -19.8303
        2016-06-16    -5.5103
        2016-07-07   -27.9703
    """
    line = AbuTLine(demean_wave, 'demean sum change wave')
    if show:
        # 计算pd_resample how='mean'只是为了_show_wave里面显示价格曲线
        xd_mean_close = pd_resample(kl_pd.close, xd_resample, how='mean')
        # 这里不使用AbuTLine.show，因为需要绘制另一个对比line，价格均线xd_mean_close
        _show_wave(demean_wave, line.high, line.mean, line.low, xd_mean_close)
        # TODO AbuTLine中添加多条线的标准对比方法，左右双轴和数据变化方式
    return line


def calc_wave_weight_abs(kl_pd, xd=21, span=3, show=True):
    """
    计算金融时间序列kl_pd的绝对波动，通过参数xd对波动进行重采样
    在默认xd＝21情况下，变成了月震荡幅度，使用ABuStatsUtil.demean对
    月震荡幅度进行去均值操作后得到技术线demean_wave，与calc_wave_abs不同，
    使用squared  * np.sign(demean_wave)放大了wave，即大的愈加大，且
    继续对squared_wave进行时间加权得到技术线形成ewm_wave，AbuTLine包裹技术线返回
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param xd: 对波动进行重采样的周期，单位天，int
    :param span: 对squared_wave进行时间加权的窗口参数，int
    :param show: 是否可视化
    :return: 返回AbuTLine对象
    """

    # 不考虑正负，只考虑波动，np.abs(kl_pd['p_change'])
    abs_pct_change = np.abs(kl_pd['p_change'])
    xd_resample = '%dD' % xd
    # 通过pd_resample重采样，使用how=sum, 即默认xd＝21情况下，变成了月震荡幅度
    change_ratio_sum = pd_resample(abs_pct_change, xd_resample, how='sum')
    # 使用ABuStatsUtil.demean进行去均值操作
    demean_wave = ABuStatsUtil.demean(change_ratio_sum)
    # 与calc_wave_abs不同，使用squared  * np.sign(demean_wave)放大了wave，即大的愈加大
    squared_wave = (demean_wave ** 2) * np.sign(demean_wave)
    # ewmd的span最后决定了一切， span默认值之对应xd默认值，xd变动, span也要变
    ewm_wave = pd_ewm_mean(squared_wave, span=span, min_periods=span, adjust=True)
    line = AbuTLine(ewm_wave, 'squared ewm wave')
    if show:
        # 计算pd_resample how='mean'只是为了_show_wave里面显示价格曲线
        xd_mean_close = pd_resample(kl_pd.close, xd_resample, how='mean')
        # 这里不使用AbuTLine.show，因为需要绘制另一个对比line，价格均线xd_mean_close
        _show_wave(ewm_wave, line.high, line.mean, line.low, xd_mean_close)
    return line


def _show_wave(wave, above, wave_mean, below, xd_mean_close):
    """
    calc_wave_abs和calc_wave_weight_abs形成技术线的可视化方法
    不使用AbuTLine.show，因为需要绘制另一个对比line，价格均线xd_mean_close，
    使用双坐标轴的方式进行可视化
    """
    fig, ax1 = plt.subplots()
    plt.plot(wave)
    plt.axhline(above, color='c')
    plt.axhline(wave_mean, color='r')
    plt.axhline(below, color='g')
    _ = plt.setp(plt.gca().get_xticklabels(), rotation=30)
    plt.legend(['wave', 'above', 'wave_mean', 'below'],
               bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

    # 采用左右两个y轴坐标显示
    # noinspection PyUnusedLocal
    ax2 = ax1.twinx()
    plt.plot(xd_mean_close, c='y')
    plt.plot(xd_mean_close, 'ro')
    plt.legend(['mean close'],
               bbox_to_anchor=(1.05, 1), loc=3, borderaxespad=0.)
    # 当时间序列太长时使用将时间显示倾斜30度角
    _ = plt.setp(plt.gca().get_xticklabels(), rotation=30)

    plt.title('wave line')
    plt.show()



# -*- encoding:utf-8 -*-
"""
    市场，数据可视化绘制模块
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import itertools
import logging
import os
from collections import Iterable
from math import pi

import bokeh.plotting as bp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from ..CoreBu import ABuEnv
from ..MarketBu import ABuSymbolPd
from ..UtilBu import ABuDateUtil
from ..UtilBu import ABuFileUtil
from ..UtilBu import ABuScalerUtil
from ..CoreBu.ABuFixes import partial

# TODO 可设置的红涨绿跌，还是绿涨红跌
__colorup__ = "red"
__colordown__ = "green"

"""预备颜色序列集，超出序列数量应使用itertools.cycle循环绘制"""
K_PLT_MAP_STYLE = [
    'b', 'c', 'g', 'k', 'm', 'r', 'y', 'w']

"""保存可视化png文件路径"""
K_SAVE_CACHE_PNG_ROOT = os.path.join(ABuEnv.g_project_data_dir, 'save_png')
"""保存可视化html文件路径"""
K_SAVE_CACHE_HTML_ROOT = os.path.join(ABuEnv.g_project_data_dir, 'save_html')

"""暂时只做全据设置，不画量只画价格"""
g_only_draw_price = False


def plot_candle_from_order(order, date_ext=120, day_sum=False, html_bk=False, save=False):
    """
    根据order绘制交易发生在金融时间序列上的位置等信息，对交易进行可视化分析时使用
    :param order: AbuOrder对象转换的pd.DataFrame对象or pd.Series对象
    :param date_ext: int对象 eg. 如交易在2015-06-01执行，如date_ext＝120，择start向前推120天，end向后推120天
    :param day_sum: 端线图 matplotlib的版本有些有bug显示不对
    :param html_bk: 使用bk绘制可交互的k图，在网页上进行交互
    :param save: 是否保存可视化结果在本地
    :return:
    """
    if not isinstance(order, (pd.DataFrame, pd.Series)) and order.shape[0] > 0:
        # order必须是pd.DataFrame对象or pd.Series对象
        raise TypeError('order must DataFrame here!!')

    is_df = isinstance(order, pd.DataFrame)
    if is_df and order.shape[0] == 1:
        # 如果是只有1行pd.DataFrame对象则变成pd.Series
        is_df = False
        # 通过iloc即变成pd.Series对象
        # noinspection PyUnresolvedReferences
        order = order.iloc[0]

    def plot_from_series(p_order, p_want_df):
        """
        根据order对交易进行可视化分析
        :param p_order: AbuOrder对象转换的pd.Series对象
        :param p_want_df: 是否返回交易买入卖出时段金融数据序列
        """

        # 确定交易对象
        target_symbol_inner = p_order['symbol']
        view_index_inner = None
        # 单子都必须有买入时间
        start = ABuDateUtil.fmt_date(p_order['buy_date'])
        # 通过date_ext确定start，即买人单子向前推date_ext天
        start = ABuDateUtil.begin_date(date_ext, date_str=start, fix=False)
        if p_order['sell_type'] != 'keep':
            # 如果有卖出，继续通过sell_date，date_ext确定end时间
            view_index_inner = [pd.to_datetime(str(p_order['buy_date'])), pd.to_datetime(str(p_order['sell_date']))]
            end = ABuDateUtil.fmt_date(p_order['sell_date'])
            # -date_ext 向前
            end = ABuDateUtil.begin_date(-date_ext, date_str=end, fix=False)
        else:
            end = None

        try:
            df = plot_candle_from_symbol(target_symbol_inner, start=start, end=end, day_sum=day_sum,
                                         html_bk=html_bk,
                                         view_index=view_index_inner, save=save)
        except Exception as e:
            logging.exception(e)
            df = None
        if p_want_df:
            return df

    if not is_df:
        return plot_from_series(order, p_want_df=True)
    else:
        # 如果是多个order，排查一下没有交易结果的
        order = order[order['result'] != 0]
        # 如果只有1个order，仍然返回plot_from_series返回的交易范围数据
        want_df = len(order) == 1
        # 多个order迭代执行plot_from_series
        keep_df = order.apply(plot_from_series, axis=1, args=(want_df,))
        if want_df:
            return keep_df


def plot_candle_from_symbol(target_symbol, n_folds=2, start=None, end=None, day_sum=False, html_bk=False,
                            view_index=None, save=False):
    """
    根据target_symbol绘制交易发生在金融时间序列上的位置等信息，对交易进行可视化分析时使用
    :param target_symbol: str对象，代表一个symbol
    :param n_folds: 请求几年的历史回测数据int
    :param start: 请求的开始日期 str对象
    :param end: 请求的结束日期 str对象
    :param day_sum: 端线图 matplotlib的版本有些有bug显示不对
    :param html_bk: 使用bk绘制可交互的k图，在网页上进行交互
    :param view_index: 需要在可视化图中重点标记的交易日信息
        eg. view_index_inner = [pd.to_datetime(str(p_order['buy_date'])), pd.to_datetime(str(p_order['sell_date']))]
    :param save: 是否保存可视化结果在本地
    :return:
    """
    # 通过make_kl_df获取需要的时间序列对象
    kl_pd = ABuSymbolPd.make_kl_df(target_symbol, n_folds=n_folds, start=start, end=end)
    if kl_pd is None or kl_pd.shape[0] == 0:
        logging.debug(target_symbol + ': has net error in data')
        return
    # 绘制获取的kl_pd对象
    plot_candle_form_klpd(kl_pd, day_sum, html_bk, view_index, save=save)

    if isinstance(view_index, Iterable) and len(view_index) == 2:
        """
            即形如[pd.to_datetime(str(p_order['buy_date'])), pd.to_datetime(str(p_order['sell_date']))]
            截取交易买入直到交易卖出这一段金融时间序列
        """
        # noinspection PyUnresolvedReferences
        return kl_pd.loc[view_index[0]:view_index[1]]

    return kl_pd


# 偏函数构成可交换可视化方法
plot_html_symbol = partial(plot_candle_from_symbol, html_bk=True)
# 偏函数构成不可交换可视化方法
plot_symbol = partial(plot_candle_from_symbol, html_bk=False)


def plot_candle_form_klpd(kl_pd, day_sum=False, html_bk=False, view_indexs=None, save=False, name=None):
    """

    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param day_sum: 端线图 matplotlib的版本有些有bug显示不对
    :param html_bk: 使用bk绘制可交互的k图，在网页上进行交互
    :param view_indexs: 需要在可视化图中重点标记的交易日信息
        eg. view_index_inner = [pd.to_datetime(str(p_order['buy_date'])), pd.to_datetime(str(p_order['sell_date']))]
    :param save: 是否保存可视化结果在本地
    :param name: 外部设置name做为可视化titile，如果不设置取kl_pd.name，即symbol name
    """
    fn = name if name else kl_pd.name if hasattr(kl_pd, 'name') else 'stock'
    plot_candle_stick(kl_pd.index, kl_pd['open'].values, kl_pd['high'].values, kl_pd['low'].values,
                      kl_pd['close'].values, kl_pd['volume'].values, view_indexs,
                      fn, day_sum, html_bk, save)


def plot_candle_stick(date, p_open, high, low, close, volume, view_index, symbol, day_sum, html_bk, save, minute=False):
    """
    展开各个k图绘制数据进行绘制
    :param date: 金融时间序列交易日时间，pd.DataFrame.index对象
    :param p_open: 金融时间序列开盘价格序列，np.array对象
    :param high: 金融时间序列最高价格序列，np.array对象
    :param low: 金融时间序列最低价格序列，np.array对象
    :param close: 金融时间序列收盘价格序列，np.array对象
    :param volume: 金融时间序列成交量序列，np.array对象
    :param view_index: 需要在可视化图中重点标记的交易日信息
        eg. view_index_inner = [pd.to_datetime(str(p_order['buy_date'])), pd.to_datetime(str(p_order['sell_date']))]
    :param symbol: symbol str对象
    :param day_sum: 端线图 matplotlib的版本有些有bug显示不对
    :param html_bk: 使用bk绘制可交互的k图，在网页上进行交互
    :param save: 是否保存可视化结果在本地
    :param minute: 是否是绘制分钟k线图
    """
    if html_bk is False:
        # 绘制不可交互的
        _do_plot_candle(date, p_open, high, low, close, volume, view_index, symbol, day_sum, save, minute)
    else:
        # 通过bk绘制可交互的
        _do_plot_candle_html(date, p_open, high, low, close, symbol, save)


def _do_plot_candle_html(date, p_open, high, low, close, symbol, save):
    """
    bk绘制可交互的k线图
    :param date: 融时间序列交易日时间，pd.DataFrame.index对象
    :param p_open: 金融时间序列开盘价格序列，np.array对象
    :param high: 金融时间序列最高价格序列，np.array对象
    :param low: 金融时间序列最低价格序列，np.array对象
    :param close: 金融时间序列收盘价格序列，np.array对象
    :param symbol: symbol str对象
    :param save: 是否保存可视化结果在本地
    """
    mids = (p_open + close) / 2
    spans = abs(close - p_open)

    inc = close > p_open
    dec = p_open > close

    w = 24 * 60 * 60 * 1000

    t_o_o_l_s = "pan,wheel_zoom,box_zoom,reset,save"

    p = bp.figure(x_axis_type="datetime", tools=t_o_o_l_s, plot_width=1280, title=symbol)
    p.xaxis.major_label_orientation = pi / 4
    p.grid.grid_line_alpha = 0.3

    p.segment(date.to_datetime(), high, date.to_datetime(), low, color="black")
    # noinspection PyUnresolvedReferences
    p.rect(date.to_datetime()[inc], mids[inc], w, spans[inc], fill_color=__colorup__, line_color=__colorup__)
    # noinspection PyUnresolvedReferences
    p.rect(date.to_datetime()[dec], mids[dec], w, spans[dec], fill_color=__colordown__, line_color=__colordown__)

    bp.show(p)
    if save:
        save_dir = os.path.join(K_SAVE_CACHE_HTML_ROOT, ABuDateUtil.current_str_date())
        html_name = os.path.join(save_dir, symbol + ".html")
        ABuFileUtil.ensure_dir(html_name)
        bp.output_file(html_name, title=symbol)


def _do_plot_candle(date, p_open, high, low, close, volume, view_index, symbol, day_sum, save, minute):
    """
    绘制不可交互的k线图
    param date: 融时间序列交易日时间，pd.DataFrame.index对象
    :param p_open: 金融时间序列开盘价格序列，np.array对象
    :param high: 金融时间序列最高价格序列，np.array对象
    :param low: 金融时间序列最低价格序列，np.array对象
    :param close: 金融时间序列收盘价格序列，np.array对象
    :param volume: 金融时间序列成交量序列，np.array对象
    :param symbol: symbol str对象
    :param day_sum: 端线图 matplotlib的版本有些有bug显示不对
    :param save: 是否保存可视化结果在本地
    :param minute: 是否是绘制分钟k线图
    """

    # 需要内部import不然每次import abupy都有warning，特别是子进程很烦人
    try:
        # noinspection PyUnresolvedReferences, PyDeprecation
        import matplotlib.finance as mpf
    except ImportError:
        # 2.2 才会有
        # noinspection PyUnresolvedReferences, PyDeprecation
        import matplotlib.mpl_finance as mpf

    if not g_only_draw_price:
        # 成交量，价格都绘制
        # noinspection PyTypeChecker
        fig, (ax1, ax2) = plt.subplots(2, sharex=True, figsize=(20, 12))
    else:
        # 只绘制价格
        fig, ax1 = plt.subplots(figsize=(6, 6))
    if day_sum:
        # 端线图绘制
        qutotes = []
        for index, (d, o, c, l, h) in enumerate(zip(date, p_open, close, low, high)):
            d = index if minute else mpf.date2num(d)
            val = (d, o, c, l, h)
            qutotes.append(val)
        # plot_day_summary_oclh接口，与mpf.candlestick_ochl不同，即数据顺序为开收低高
        mpf.plot_day_summary_oclh(ax1, qutotes, ticksize=5, colorup=__colorup__, colordown=__colordown__)
    else:
        # k线图绘制
        qutotes = []
        for index, (d, o, c, h, l) in enumerate(zip(date, p_open, close, high, low)):
            d = index if minute else mpf.date2num(d)
            val = (d, o, c, h, l)
            qutotes.append(val)
        # mpf.candlestick_ochl即数据顺序为开收高低
        mpf.candlestick_ochl(ax1, qutotes, width=0.6, colorup=__colorup__, colordown=__colordown__)

    if not g_only_draw_price:
        # 开始绘制成交量
        ax1.set_title(symbol)
        ax1.set_ylabel('ochl')
        ax1.grid(True)
        if not minute:
            ax1.xaxis_date()
        if view_index is not None:
            # 开始绘制买入交易日，卖出交易日，重点突出的点位
            e_list = date.tolist()
            # itertools.cycle循环使用备选的颜色
            for v, csColor in zip(view_index, itertools.cycle(K_PLT_MAP_STYLE)):
                try:
                    v_ind = e_list.index(v)
                except Exception as e:
                    # logging.exception(e)
                    logging.debug(e)
                    # 向前倒一个
                    v_ind = len(close) - 1
                label = symbol + ': ' + str(date[v_ind])
                ax1.plot(v, close[v_ind], 'ro', markersize=12, markeredgewidth=4.5,
                         markerfacecolor='None', markeredgecolor=csColor, label=label)

                # 因为candlestick_ochl 不能label了，所以使用下面的显示文字
                # noinspection PyUnboundLocalVariable
                ax2.plot(v, 0, 'ro', markersize=12, markeredgewidth=0.5,
                         markerfacecolor='None', markeredgecolor=csColor, label=label)
            plt.legend(loc='best')

        # 成交量柱子颜色，收盘价格 > 开盘，即红色
        # noinspection PyTypeChecker
        bar_red = np.where(close >= p_open, volume, 0)
        # 成交量柱子颜色，开盘价格 > 收盘。即绿色
        # noinspection PyTypeChecker
        bar_green = np.where(p_open > close, volume, 0)

        date = date if not minute else np.arange(0, len(date))
        ax2.bar(date, bar_red, facecolor=__colorup__)
        ax2.bar(date, bar_green, facecolor=__colordown__)

        ax2.set_ylabel('volume')
        ax2.grid(True)
        ax2.autoscale_view()
        plt.setp(plt.gca().get_xticklabels(), rotation=30)
    else:
        ax1.grid(False)

    if save:
        # 保存可视化结果在本地
        from pylab import savefig
        save_dir = os.path.join(K_SAVE_CACHE_PNG_ROOT, ABuDateUtil.current_str_date())
        png_dir = os.path.join(save_dir, symbol)
        ABuFileUtil.ensure_dir(png_dir)
        r_cnt = 0
        while True:
            png_name = '{}{}.png'.format(png_dir, '' if r_cnt == 0 else '-{}'.format(r_cnt))
            if not ABuFileUtil.file_exist(png_name):
                break
            r_cnt += 1
        # noinspection PyUnboundLocalVariable
        savefig(png_name)
        fig.clf()
        plt.close(fig)
    else:
        """
            save 了就不show了
        """
        plt.show()


def save_dir_name(html=False):
    """
    外部获取缓存文件夹的绝对路径
    :param html: 是否缓存为html文件
    """
    r_dir = K_SAVE_CACHE_HTML_ROOT if html else K_SAVE_CACHE_PNG_ROOT
    save_dir = os.path.join(r_dir, ABuDateUtil.current_str_date())
    return save_dir


def plot_simple_multi_stock(multi_kl_pd):
    """
    将多个金融时间序列收盘价格缩放到一个价格水平后，可视化价格变动
    :param multi_kl_pd: 可迭代的序列，元素为金融时间序列
    """
    rg_ret = ABuScalerUtil.scaler_matrix([kl_pd.close for kl_pd in multi_kl_pd])
    for i, kl_pd in enumerate(multi_kl_pd):
        plt.plot(kl_pd.index, rg_ret[i])
    plt.show()


def plot_simple_two_stock(two_stcok_dict):
    """
    将两个金融时间序列收盘价格缩放到一个价格水平后，可视化价格变动
    :param two_stcok_dict: 字典形式，key将做为lable进行可视化使用，元素为金融时间序列
    """
    if not isinstance(two_stcok_dict, dict) or len(two_stcok_dict) != 2:
        print('two_stcok_dict type must dict! or len(two_stcok_dict) != 2')
        return

    label_arr = [s_name for s_name in two_stcok_dict.keys()]
    x, y = ABuScalerUtil.scaler_xy(two_stcok_dict[label_arr[0]].close, two_stcok_dict[label_arr[1]].close)
    plt.plot(x, label=label_arr[0])
    plt.plot(y, label=label_arr[1])
    plt.legend(loc=2)
    plt.show()



# coding=utf-8
"""
    数据源解析模块以及示例内置数据源的解析类实现
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging

import numpy as np
import pandas as pd

from .ABuSymbol import EMarketTargetType
from ..CoreBu.ABuFixes import six
from ..UtilBu import ABuDateUtil

__author__ = '阿布'
__weixin__ = 'abu_quant'


def del_columns(df, columns):
    """
    从df中删除参数columns指定的整列数据
    :param df: 金融时间序列切割pd.DataFrame对象
    :param columns: 可迭代的字符序列，代表需要删除的指定列
    :return:
    """
    old_c = df.columns.tolist()
    for col in filter(lambda x: x in old_c, columns):
        df.drop(col, axis=1, inplace=True)


class AbuDataParseWrap(object):
    """
        做为类装饰器封装替换解析数据统一操作，装饰替换init
    """

    def __call__(self, cls):
        """只做为数据源解析类的装饰器，统一封装通用的数据解析规范及流程"""
        if isinstance(cls, six.class_types):
            # 只做为类装饰器使用
            init = cls.__init__

            def wrapped(*args, **kwargs):
                try:
                    # 拿出被装饰的self对象
                    warp_self = args[0]
                    warp_self.df = None
                    # 调用原始init
                    init(*args, **kwargs)
                    symbol = args[1]
                    # 开始数据解析
                    self._gen_warp_df(warp_self, symbol)
                except Exception as e:
                    logging.exception(e)

            # 使用wrapped替换原始init
            cls.__init__ = wrapped

            wrapped.__name__ = '__init__'
            # 将原始的init赋予deprecated_original，必须要使用这个属性名字，在其它地方，如AbuParamBase会寻找原始方法找它
            wrapped.deprecated_original = init
            return cls
        else:
            raise TypeError('AbuDataParseWrap just for class warp')

    # noinspection PyMethodMayBeStatic
    def _gen_warp_df(self, warp_self, symbol):
        """
        封装通用的数据解析规范及流程
        :param warp_self: 被封装类init中使用的self对象
        :param symbol: 请求的symbol str对象
        :return:
        """

        # 规范原始init函数中必须为类添加了如下属性
        must_col = ['open', 'close', 'high', 'low', 'volume', 'date']
        # 检测所有的属性都有
        all_has = all([hasattr(warp_self, col) for col in must_col])
        # raise RuntimeError('df.columns must have |date|open|close|high|volume| ')
        if all_has:
            # 将时间序列转换为pd时间
            dates_pd = pd.to_datetime(warp_self.date)
            # 构建df，index使用dates_pd
            warp_self.df = pd.DataFrame(index=dates_pd)
            for col in must_col:
                # 所以必须有的类属性序列设置给df的列
                warp_self.df[col] = getattr(warp_self, col)

            # 从收盘价格序列shift出昨收价格序列
            warp_self.df['pre_close'] = warp_self.df['close'].shift(1)
            warp_self.df['pre_close'].fillna(warp_self.df['open'], axis=0, inplace=True)
            # 添加日期int列
            warp_self.df['date'] = warp_self.df['date'].apply(lambda x: ABuDateUtil.date_str_to_int(str(x)))
            # 添加周几列date_week，值为0-4，分别代表周一到周五
            warp_self.df['date_week'] = warp_self.df['date'].apply(
                lambda x: ABuDateUtil.week_of_date(str(x), '%Y%m%d'))

            # 类型转换
            warp_self.df['close'] = warp_self.df['close'].astype(float)
            warp_self.df['high'] = warp_self.df['high'].astype(float)
            warp_self.df['low'] = warp_self.df['low'].astype(float)
            warp_self.df['open'] = warp_self.df['open'].astype(float)
            warp_self.df['volume'] = warp_self.df['volume'].astype(float)
            warp_self.df['volume'] = warp_self.df['volume'].astype(np.int64)
            warp_self.df['date'] = warp_self.df['date'].astype(int)
            warp_self.df['pre_close'] = warp_self.df['pre_close'].astype(float)
            # 不使用df['close'].pct_change计算
            # noinspection PyTypeChecker
            warp_self.df['p_change'] = np.where(warp_self.df['pre_close'] == 0, 0,
                                                (warp_self.df['close'] - warp_self.df['pre_close']) / warp_self.df[
                                                    'pre_close'] * 100)

            warp_self.df['p_change'] = warp_self.df['p_change'].apply(lambda x: round(x, 3))
            # 给df加上name
            warp_self.df.name = symbol


@AbuDataParseWrap()
class TXParser(object):
    """tx数据源解析类，被类装饰器AbuDataParseWrap装饰"""

    def __init__(self, symbol, sub_market, json_dict):
        """
        :param symbol: 请求的symbol str对象
        :param sub_market: 子市场（交易所）类型
        :param json_dict: 请求返回的json数据
        """
        if json_dict['code'] == 0:
            if symbol.market == EMarketTargetType.E_MARKET_TARGET_US:
                data = json_dict['data'][symbol.value + sub_market]
            else:
                data = json_dict['data'][symbol.value]

            if 'qfqday' in data.keys():
                data = data['qfqday']
            else:
                data = data['day']

            # 为AbuDataParseWrap准备类必须的属性序列
            if len(data) > 0:
                # 时间日期序列，时间格式为2017-07-26格式字符串
                self.date = [item[0] for item in data]
                # 开盘价格序列
                self.open = [item[1] for item in data]
                # 收盘价格序列
                self.close = [item[2] for item in data]
                # 最高价格序列
                self.high = [item[3] for item in data]
                # 最低价格序列
                self.low = [item[4] for item in data]
                # 成交量序列
                self.volume = [item[5] for item in data]


@AbuDataParseWrap()
class NTParser(object):
    """nt数据源解析类，被类装饰器AbuDataParseWrap装饰"""

    # noinspection PyUnusedLocal
    def __init__(self, symbol, json_dict):
        """
        :param symbol: 请求的symbol str对象
        :param json_dict: 请求返回的json数据
        """
        data = json_dict['data']
        # 为AbuDataParseWrap准备类必须的属性序列
        if len(data) > 0:
            # 时间日期序列
            self.date = [item[0] for item in data]
            # 开盘价格序列
            self.open = [item[1] for item in data]
            # 收盘价格序列
            self.close = [item[2] for item in data]
            # 最高价格序列
            self.high = [item[3] for item in data]
            # 最低价格序列
            self.low = [item[4] for item in data]
            # 成交量序列
            self.volume = [item[5] for item in data]


@AbuDataParseWrap()
class SNUSParser(object):
    """snus数据源解析类，被类装饰器AbuDataParseWrap装饰"""

    # noinspection PyUnusedLocal
    def __init__(self, symbol, json_dict):
        """
        :param symbol: 请求的symbol str对象
        :param json_dict: 请求返回的json数据
        """
        data = json_dict
        # 为AbuDataParseWrap准备类必须的属性序列
        if len(data) > 0:
            # 时间日期序列
            self.date = [item['d'] for item in data]
            # 开盘价格序列
            self.open = [item['o'] for item in data]
            # 收盘价格序列
            self.close = [item['c'] for item in data]
            # 最高价格序列
            self.high = [item['h'] for item in data]
            # 最低价格序列
            self.low = [item['l'] for item in data]
            # 成交量序列
            self.volume = [item['v'] for item in data]


@AbuDataParseWrap()
class SNFuturesParser(object):
    """示例期货数据源解析类，被类装饰器AbuDataParseWrap装饰"""

    # noinspection PyUnusedLocal
    def __init__(self, symbol, json_dict):
        """
        :param symbol: 请求的symbol str对象
        :param json_dict: 请求返回的json数据
        """
        data = json_dict
        # 为AbuDataParseWrap准备类必须的属性序列
        if len(data) > 0:
            # 时间日期序列
            self.date = [item[0] for item in data]
            # 开盘价格序列
            self.open = [item[1] for item in data]
            # 最高价格序列
            self.high = [item[2] for item in data]
            # 最低价格序列
            self.low = [item[3] for item in data]
            # 收盘价格序列
            self.close = [item[4] for item in data]
            # 成交量序列
            self.volume = [item[5] for item in data]


@AbuDataParseWrap()
class SNFuturesGBParser(object):
    """示例国际期货数据源解析类，被类装饰器AbuDataParseWrap装饰"""
    # noinspection PyUnusedLocal
    def __init__(self, symbol, json_dict):
        """
        :param symbol: 请求的symbol str对象
        :param json_dict: 请求返回的json或者dict数据
        """
        data = json_dict
        # 为AbuDataParseWrap准备类必须的属性序列
        if len(data) > 0:
            # 时间日期序列
            self.date = [item['date'] for item in data]
            # 开盘价格序列
            self.open = [item['open'] for item in data]
            # 最高价格序列
            self.high = [item['high'] for item in data]
            # 最低价格序列
            self.low = [item['low'] for item in data]
            # 收盘价格序列
            self.close = [item['close'] for item in data]
            # 成交量序列
            self.volume = [item['volume'] for item in data]


@AbuDataParseWrap()
class HBTCParser(object):
    """示例币类市场数据源解析类，被类装饰器AbuDataParseWrap装饰"""

    # noinspection PyUnusedLocal
    def __init__(self, symbol, json_dict):
        """
        :param symbol: 请求的symbol str对象
        :param json_dict: 请求返回的json数据
        """

        data = json_dict
        # 为AbuDataParseWrap准备类必须的属性序列
        if len(data) > 0:
            # 时间日期序列
            self.date = [item[0] for item in data]
            # 开盘价格序列
            self.open = [item[1] for item in data]
            # 最高价格序列
            self.high = [item[2] for item in data]
            # 最低价格序列
            self.low = [item[3] for item in data]
            # 收盘价格序列
            self.close = [item[4] for item in data]
            # 成交量序列
            self.volume = [item[5] for item in data]

            # 时间日期进行格式转化，转化为如2017-07-26格式字符串
            self.date = list(map(lambda date: ABuDateUtil.fmt_date(date), self.date))


class BDParser(object):
    """bd数据源解析类"""

    data_keys = ['data', 'dataMash']
    s_calc_dm = True

    def __init__(self, symbol, json_dict):
        """
        没有使用AbuDataParseWrap装饰类，保留一个原始的解析流程类，
        其它的解析类都使用AbuDataParseWrap装饰类，解析过程不做多注解，
        详阅读AbuDataParseWrap的实现
        :param symbol: 请求的symbol str对象
        :param json_dict: 请求返回的json数据
        """
        try:
            if BDParser.data_keys[0] in json_dict.keys():
                self.data = json_dict[BDParser.data_keys[0]][::-1]
            elif BDParser.data_keys[1] in json_dict.keys():
                self.data = json_dict[BDParser.data_keys[1]][::-1]
            else:
                raise ValueError('content not json format')

            dates = [mash['date'] for mash in self.data]
            klines = [mash['kline'] for mash in self.data]

            self.df = None
            if len(klines) > 0 and len(dates) > 0:
                dates_fmt = list(map(lambda date: ABuDateUtil.fmt_date(date), dates))
                dates_pd = pd.to_datetime(dates_fmt)

                self.df = pd.DataFrame(klines, index=dates_pd)
                self.df['date'] = dates
                self.df['date_week'] = self.df['date'].apply(lambda x: ABuDateUtil.week_of_date(str(x), '%Y%m%d'))

                self.df['close'] = self.df['close'].astype(float)
                self.df['high'] = self.df['high'].astype(float)
                self.df['low'] = self.df['low'].astype(float)
                self.df['open'] = self.df['open'].astype(float)
                self.df['volume'] = self.df['volume'].astype(np.int64)
                self.df['date'] = self.df['date'].astype(int)
                self.df['netChangeRatio'] = self.df['netChangeRatio'].map(lambda x: x[:-1]).astype(float)
                self.df['preClose'] = self.df['preClose'].astype(float)
                self.df.rename(columns={'preClose': 'pre_close', 'netChangeRatio': 'p_change'}, inplace=True)
                del_columns(self.df, ['amount'])
                if BDParser.s_calc_dm:
                    self.df['pre_close'] = self.df['close'].shift(1)
                    self.df['pre_close'].fillna(self.df['open'], axis=0, inplace=True)
                    # 不使用df['close'].pct_change计算
                    # noinspection PyTypeChecker
                    self.df['p_change'] = np.where(self.df['pre_close'] == 0, 0,
                                                   (self.df['close'] - self.df['pre_close']) / self.df[
                                                       'pre_close'] * 100)
                    self.df['p_change'] = self.df['p_change'].apply(lambda x: round(x, 3))

                self.df.name = symbol

        except Exception as e:
            logging.exception(e)



# -*- encoding:utf-8 -*-
from __future__ import absolute_import

# noinspection all
from . import ABuPositionBase as position
# noinspection all
from . import ABuAtrPosition as atr
# noinspection all
from . import ABuKellyPosition as kelly



from __future__ import absolute_import

from .ABuFactorBuyBase import AbuFactorBuyBase, AbuFactorBuyXD, AbuFactorBuyTD, BuyCallMixin, BuyPutMixin
from .ABuFactorBuyBreak import AbuFactorBuyBreak, AbuFactorBuyPutBreak
from .ABuFactorBuyWD import AbuFactorBuyWD
from .ABuFactorBuyDemo import AbuSDBreak, AbuTwoDayBuy, AbuWeekMonthBuy, AbuFactorBuyBreakUmpDemo
from .ABuFactorBuyDemo import AbuFactorBuyBreakReocrdHitDemo, AbuFactorBuyBreakHitPredictDemo
from .ABuFactorBuyDM import AbuDoubleMaBuy

__all__ = [
    'AbuFactorBuyBase',
    'AbuFactorBuyXD',
    'AbuFactorBuyTD',
    'BuyCallMixin',
    'BuyPutMixin',
    'AbuFactorBuyBreak',
    'AbuFactorBuyWD',
    'AbuFactorBuyPutBreak',
    'AbuFactorBuyBreakUmpDemo',
    'AbuFactorBuyBreakReocrdHitDemo',
    'AbuFactorBuyBreakHitPredictDemo',
    'AbuSDBreak',
    'AbuTwoDayBuy',
    'AbuWeekMonthBuy',
    'AbuDoubleMaBuy'
]



from __future__ import absolute_import

from .ABuProcessor import arg_process, return_process
from .ABuChecker import FuncChecker, ArgChecker, ReturnChecker

__all__ = [
    'arg_process',
    'return_process',
    'FuncChecker',
    'ArgChecker',
    'ReturnChecker',
]



from __future__ import absolute_import

from . import ABuCorrcoef
from . import ABuSimilar
from .ABuCorrcoef import ECoreCorrType
from .ABuSimilar import find_similar_with_se, find_similar_with_folds, find_similar_with_cnt

__all__ = [
    'ABuSimilar',
    'ABuSimilarDrawing',
    'ABuCorrcoef',
    'ECoreCorrType',
    'find_similar_with_se',
    'find_similar_with_folds',
    'find_similar_with_cnt'
]



# -*- encoding:utf-8 -*-
"""示例ump边裁特征价格模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..TradeBu.ABuMLFeature import AbuFeaturePrice
from ..MLBu.ABuMLPd import AbuMLPd
from .ABuUmpBase import ump_edge_make_xy, BuyUmpMixin
from .ABuUmpEdgeBase import AbuUmpEdgeBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuUmpEdgePrice(AbuUmpEdgeBase, BuyUmpMixin):
    """边裁价格特征类，AbuUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpPriceFiter(AbuMLPd):
        """
            内部类，只需要实现make_xy，且使用ump_edge_make_xy装饰

            在边裁__init__中通过：
                    self.fiter_cls = self.get_fiter_class()
                    self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
            构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
        """

        @ump_edge_make_xy
        def make_xy(self, **kwarg):
            """
             make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
             order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
             make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

            :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                          kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
            :return: self.order_has_ret中使用filter选取特征列形成price_df
                     ump_edge_make_xy装饰器在make_xy返回price_df后做转换matrix，形成x，y等工作
            """

            filter_list = ['profit', 'profit_cg']
            cols = AbuFeaturePrice().get_feature_ump_keys(ump_cls=AbuUmpEdgePrice)
            # ['profit', 'profit_cg', 'buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
            filter_list.extend(cols)

            # noinspection PyUnresolvedReferences
            price_df = self.order_has_ret.filter(filter_list)
            """
                eg: price_df
                               profit  profit_cg  buy_price_rank120  buy_price_rank90  \
                2014-09-24  -22618.04    -0.0566              1.000             1.000
                2014-10-24  -29690.28    -0.0742              1.000             1.000
                2014-10-29   18959.19     0.0542              1.000             1.000
                2014-10-29  148209.36     0.5022              0.925             0.900
                2014-10-29   24867.60     0.0952              0.962             0.950
                2014-10-29   18959.19     0.0542              1.000             1.000
                2014-11-03    1250.80     0.0045              1.000             1.000
                2014-11-11   59888.21     0.1857              0.954             0.939
                2014-11-12   -3578.78    -0.0140              0.475             0.522
                2014-11-26  -29085.19    -0.0946              0.642             0.733
                ...               ...        ...                ...               ...
                2016-03-14   16220.57     0.0559              0.617             0.500
                2016-03-14  -25328.12    -0.1218              0.683             0.589
                2016-03-30  -29858.44    -0.0863              0.658             0.667
                2016-04-04    5373.76     0.0244              0.400             0.511
                2016-04-13  -28044.40    -0.1159              0.567             0.722
                2016-04-14  -18645.93    -0.0467              0.875             0.878
                2016-04-15  -32484.79    -0.1149              0.775             0.733
                2016-04-15  -32484.79    -0.1149              0.775             0.733
                2016-04-29     290.96     0.0007              1.000             1.000
                2016-04-29     290.96     0.0007              1.000             1.000

                            buy_price_rank60  buy_price_rank252
                2014-09-24             1.000              1.000
                2014-10-24             1.000              1.000
                2014-10-29             1.000              1.000
                2014-10-29             0.883              0.750
                2014-10-29             0.925              0.982
                2014-10-29             1.000              1.000
                2014-11-03             1.000              1.000
                2014-11-11             0.992              0.808
                2014-11-12             0.783              0.560
                2014-11-26             1.000              0.762
                ...                      ...                ...
                2016-03-14             0.750              0.444
                2016-03-14             0.850              0.623
                2016-03-30             1.000              0.536
                2016-04-04             0.767              0.190
                2016-04-13             1.000              0.270
                2016-04-14             0.967              0.940
                2016-04-15             1.000              0.631
                2016-04-15             1.000              0.631
                2016-04-29             1.000              1.000
                2016-04-29             1.000              1.000
            """

            return price_df

    def get_predict_col(self):
        """
        边裁价格特征keys：['buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        :return: ['buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        """
        return AbuFeaturePrice().get_feature_ump_keys(ump_cls=AbuUmpEdgePrice)

    def get_fiter_class(self):
        """
        边裁价格特征返回的AbuMLPd子类：AbuUmpEdgePrice.UmpPriceFiter
        :return: AbuUmpEdgePrice.UmpPriceFiter
        """
        return AbuUmpEdgePrice.UmpPriceFiter

    @classmethod
    def class_unique_id(cls):
        """
         具体ump类关键字唯一名称，类方法：return 'price_edge'
         主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
         具体使用见ABuUmpManager中extend_ump_block方法
        """
        return 'price_edge'



# -*- encoding:utf-8 -*-
"""封装grid search相关操作模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import logging

import matplotlib.pyplot as plt
import numpy as np
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.metrics import explained_variance_score, make_scorer

from ..CoreBu.ABuFixes import signature, six
from ..CoreBu.ABuFixes import GridSearchCV

__author__ = '阿布'
__weixin__ = 'abu_quant'

__all__ = [
    'grid_search_init_kwargs',
    'grid_search_mul_init_kwargs',
    'grid_search_init_n_estimators',
    'grid_search_init_n_components',
    'grid_search_init_max_depth',
    'grid_search_init_n_neighbors'
]


def _scoring_grid(estimator, scoring):
    """
    只针对有监督学习过滤无监督学习，对scoring未赋予的情况根据
    学习器分类器使用accuracy进行度量，回归器使用可释方差值explained_variance_score，
    使用make_scorer对函数进行score封装

    :param estimator: 学习器对象
    :param scoring: 度量使用的方法，未赋予的情况根据
                    学习器分类器使用accuracy进行度量，回归器使用explained_variance_score进行度量
    :return: scoring
    """

    if not isinstance(estimator, (ClassifierMixin, RegressorMixin)):
        logging.info('only support supervised learning')
        # TODO 无监督学习的scoring度量以及GridSearchCV
        return None

    if scoring is None:
        if isinstance(estimator, ClassifierMixin):
            # 分类器使用accuracy
            return 'accuracy'
        elif isinstance(estimator, RegressorMixin):
            # 回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
            """
                make_scorer中通过greater_is_better对返回值进行正负分配
                eg: sign = 1 if greater_is_better else -1
            """
            return make_scorer(explained_variance_score, greater_is_better=True)
        return None
    return scoring


def grid_search_init_kwargs(estimator, x, y, param_name, param_range, cv=10, n_jobs=-1, scoring=None, show=True):
    """
    对GridSearchCV进行封装，对单个目标关键字参数进行grid search最优参数搜寻
        eg：'n_estimators'， 'max_depth'
        eg：param_range=np.arange(100, 500, 50))对最优参数进行寻找

    eg:
        from abupy import AbuML, ml
        ttn_abu = AbuML.create_test_more_fiter()
        ttn_abu.estimator.random_forest_classifier()
        ml.grid_search_init_kwargs(ttn_abu.estimator.clf, ttn_abu.x, ttn_abu.y,
                           param_name='n_estimators', param_range=np.arange(100, 500, 50))

        可找到n_estimators参数最优为：(0.81930415263748602, {'n_estimators': 300})

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param param_name: 做为grid的目标关键字参数，eg：'n_estimators'， 'max_depth'
    :param param_range: 做为grid的目标关键字参数的grid序列，eg：param_range=np.arange(100, 500, 50))
    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg：(0.81930415263748602, {'n_estimators': 300})
    """

    if not isinstance(param_name, six.string_types):
        # param_name参数需要是字符串类型
        logging.info('param_name is str, not {}, eg: \'n_estimators\''.format(param_name))
        return None, None

    # 根据分类回归得到非None的score
    scoring = _scoring_grid(estimator, scoring)
    if scoring is None:
        # 如果_scoring_grid返回的结果仍然是None, 说明无监督学习，暂时不支持
        return None, None

    # 获取学习器的init函数，使用getattr
    init = getattr(estimator.__class__.__init__, 'deprecated_original', estimator.__class__.__init__)
    # 获取函数签名
    init_signature = signature(init)
    """
        eg：init_signature
            ['self', 'base_estimator', 'n_estimators', 'max_samples', 'max_features', 'bootstrap',
            'bootstrap_features', 'oob_score', 'warm_start', 'n_jobs', 'random_state', 'verbose']
    """

    if param_name not in init_signature.parameters.keys():
        # 如果需要grid的参数param_name不在init函数签名中，打log，返回
        logging.info('check init signature {} not in **kwargs\ninit_signature:{}'.format(
            param_name, init_signature.parameters.keys()))
        return None, None

    param_grid = {param_name: param_range}
    grid = GridSearchCV(estimator, param_grid, cv=cv, scoring=scoring, n_jobs=n_jobs)
    print('start grid search please wait...')
    grid.fit(x, y)

    if show:
        if hasattr(grid, 'cv_results_'):
            # 0.18之后的版本有cv_results_，拿出每一次的训练score的mean形成grid_scores
            cv_results = grid.cv_results_
            grid_scores = cv_results['mean_test_score']
        else:
            # 0.18之前的版本
            cv_results = grid.grid_scores_
            grid_scores = [result.mean_validation_score for result in cv_results]
            """
                cv_results中每一个元素为_CVScoreTuple namedtuple对象，如下所示：

                class _CVScoreTuple (namedtuple('_CVScoreTuple',
                                    ('parameters',
                                     'mean_validation_score',
                                     'cv_validation_scores'))):
            """

        # FIXME 这里假定了所有param_range的元素类型都是数值类型，需要判定，并且根据情况是否需要排序
        plt.plot(param_range, grid_scores)
        # 把最好的红圈标记出来
        plt.plot(grid.best_params_[param_name], grid.best_score_, 'ro', markersize=12, markeredgewidth=1.5,
                 markerfacecolor='None', markeredgecolor='r')
        plt.title('search {}, best is {}'.format(param_name, grid.best_params_[param_name]))
        plt.show()
    return grid.best_score_, grid.best_params_


def grid_search_mul_init_kwargs(estimator, x, y, param_grid, cv=10, n_jobs=-1, scoring=None, show=True):
    """
    对GridSearchCV进行封装，对多个目标关键字参数进行grid search最优参数搜寻

    eg:
        from abupy import AbuML, ml
        ttn_abu = AbuML.create_test_more_fiter()
        ttn_abu.estimator.random_forest_classifier()

        param_grid = {'max_depth': np.arange(2, 5), 'n_estimators': np.arange(100, 300, 50)}
        ml.grid_search_mul_init_kwargs(ttn_abu.estimator.clf, ttn_abu.x, ttn_abu.y, param_grid=param_grid)

        out: (0.81593714927048255, {'max_depth': 4, 'n_estimators': 250})

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param param_grid: eg：param_grid = {'max_depth': np.arange(2, 5), 'n_estimators': np.arange(100, 300, 50)}
    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.81593714927048255, {'max_depth': 4, 'n_estimators': 250})
    """

    if not isinstance(param_grid, dict):
        # param_grid参数是dict对象
        logging.info('param_grid is dict object, not {}}'.format(param_grid))
        return None, None

    # 根据分类回归得到非None的score
    scoring = _scoring_grid(estimator, scoring)
    if scoring is None:
        # 如果_scoring_grid返回的结果仍然是None, 说明无监督学习，暂时不支持
        return None, None

    for param_name in param_grid.keys():
        # 迭代每一个key，即每一个关键字参数，看init方法中的签名是否存在该关键字

        # 获取学习器的init函数，使用getattr
        init = getattr(estimator.__class__.__init__, 'deprecated_original', estimator.__class__.__init__)
        # 获取函数签名
        init_signature = signature(init)
        if param_name not in init_signature.parameters.keys():
            # 如果需要grid的参数param_name不在init函数签名中，打log，返回
            logging.info('check init signature {} not in **kwargs\ninit_signature:{}'.format(
                param_name, init_signature.parameters.keys()))
            return None, None

    grid = GridSearchCV(estimator, param_grid, cv=cv, scoring=scoring, n_jobs=n_jobs)
    print('start grid search please wait...')
    grid.fit(x, y)

    if show:
        if hasattr(grid, 'cv_results_'):
            # 0.18之后的版本有cv_results_，拿出每一次的训练score的mean形成grid_scores
            cv_results = grid.cv_results_
            grid_scores = cv_results['mean_test_score']
            """
                eg：grid_scores
                [0.77890011223344557, 0.79349046015712688, 0.77553310886644222, 0.77441077441077444,
                0.80920314253647585, 0.80920314253647585, 0.80808080808080807, 0.81032547699214363,
                0.80695847362514028, 0.81144781144781142, 0.80471380471380471, 0.81593714927048255]
            """
            grid_params = cv_results['params']
            """
                eg：grid_params
                [{'max_depth': 2, 'n_estimators': 100}, {'max_depth': 2, 'n_estimators': 150},
                {'max_depth': 2, 'n_estimators': 200}, {'max_depth': 2, 'n_estimators': 250},
                {'max_depth': 3, 'n_estimators': 100}, {'max_depth': 3, 'n_estimators': 150},
                {'max_depth': 3, 'n_estimators': 200}, {'max_depth': 3, 'n_estimators': 250},
                {'max_depth': 4, 'n_estimators': 100}, {'max_depth': 4, 'n_estimators': 150},
                {'max_depth': 4, 'n_estimators': 200}, {'max_depth': 4, 'n_estimators': 250}]
            """
        else:
            cv_results = grid.grid_scores_
            """
                cv_results中每一个元素为_CVScoreTuple namedtuple对象，如下所示：

                class _CVScoreTuple (namedtuple('_CVScoreTuple',
                                    ('parameters',
                                     'mean_validation_score',
                                     'cv_validation_scores'))):
            """
            grid_scores = [result.mean_validation_score for result in cv_results]
            grid_params = [result.parameters for result in cv_results]

        # 与grid_search_init_kwargs不同可视化grid_scores绘制曲线y，x只使用index
        plt.plot(grid_scores)
        cmap = plt.get_cmap('jet', len(grid_scores))
        cmap.set_under('gray')
        for grid_index in np.arange(0, len(grid_scores)):
            # 迭代每一个分数，绘制点在曲线上根据分数用颜色区分，使用label进行标注
            plt.scatter(grid_index, grid_scores[grid_index], s=50, cmap=cmap,
                        vmin=np.min(grid_scores),
                        vmax=np.max(grid_scores),
                        label='{}: {:.2f}'.format(grid_params[grid_index], grid_scores[grid_index]))
        plt.title('best params is {}'.format(grid.best_params_))
        # 将label标注文字绘制在外面
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.show()
    return grid.best_score_, grid.best_params_


def grid_search_init_n_estimators(estimator, x, y, n_estimators_range=None, cv=10, n_jobs=-1,
                                  scoring=None, show=True):
    """
    封装grid search特定的'n_estimators'关键字参数最优搜索，
    为AbuMLCreater中_estimators_prarms_best提供callback函数，

    具体阅读
            AbuMLCreater._estimators_prarms_best()
            + AbuMLCreater.random_forest_classifier_best()

    eg:
        from abupy import AbuML, ml
        ttn_abu = AbuML.create_test_more_fiter()
        ttn_abu.estimator.random_forest_classifier()
        ml.grid_search_init_n_estimators(ttn_abu.estimator.clf, ttn_abu.x, ttn_abu.y)

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_estimators_range: 默认None, None则会使用 n_estimators_range = np.arange(50, 500, 10)
    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.82154882154882158, {'n_estimators': 310})
    """

    if n_estimators_range is None:
        n_estimators_range = np.arange(50, 500, 10)

    return grid_search_init_kwargs(estimator, x, y, 'n_estimators', n_estimators_range,
                                   cv=cv, n_jobs=n_jobs, scoring=scoring, show=show)


def grid_search_init_max_depth(estimator, x, y, max_depth_range=None, cv=10, n_jobs=-1,
                               scoring=None, show=True):
    """
    封装grid search特定的'n_components'关键字参数最优搜索，
    为AbuMLCreater中_estimators_prarms_best提供callback函数

    具体阅读
            AbuMLCreater._estimators_prarms_best()
            + AbuMLCreater.decision_tree_classifier_best()

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param max_depth_range: 默认None, None则会使用:
            max_depth_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.82154882154882158, {'max_depth': 3})
    """

    if max_depth_range is None:
        max_depth_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    return grid_search_init_kwargs(estimator, x, y, 'max_depth', max_depth_range,
                                   cv=cv, n_jobs=n_jobs, scoring=scoring, show=show)


def grid_search_init_n_neighbors(estimator, x, y, n_neighbors_range=None, cv=10, n_jobs=-1,
                                 scoring=None, show=True):
    """
    封装grid search特定的'n_components'关键字参数最优搜索，
    为AbuMLCreater中_estimators_prarms_best提供callback函数

    具体阅读
            AbuMLCreater._estimators_prarms_best()
            + AbuMLCreater.knn_classifier_best()

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_neighbors_range: 默认None, None则会使用:
            n_estimators_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.82154882154882158, {'n_components': 10})
    """

    if n_neighbors_range is None:
        # 邻居投票者控制在1-np.minimum(26, 总数的1/3）
        n_neighbors_range = np.arange(1, np.minimum(26, int(x.shape[0] / 3)), 1)

    return grid_search_init_kwargs(estimator, x, y, 'n_neighbors', n_neighbors_range,
                                   cv=cv, n_jobs=n_jobs, scoring=scoring, show=show)


def grid_search_init_n_components(estimator, x, y, n_components_range=None, cv=10, n_jobs=-1,
                                  scoring=None, show=True):
    """
    封装grid search特定的'n_components'关键字参数最优搜索，
    为AbuMLCreater中_estimators_prarms_best提供callback函数，
    具体阅读AbuMLCreater._estimators_prarms_best()

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_components_range: 默认None, None则会使用:
            n_estimators_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，回归器使用
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.82154882154882158, {'n_components': 10})
    """
    if n_components_range is None:
        n_components_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    return grid_search_init_kwargs(estimator, x, y, 'n_components', n_components_range,
                                   cv=cv, n_jobs=n_jobs, scoring=scoring, show=show)



# -*- encoding:utf-8 -*-
"""比特币度量模块"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from ..MetricsBu.ABuMetricsFutures import AbuMetricsFutures

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuMetricsTC(AbuMetricsFutures):
    """比特币，莱特币等币类型度量，自扩张使用，暂时继承AbuMetricsFutures，即不涉及benchmark，user可继承扩展需求"""



# -*- encoding:utf-8 -*-
from __future__ import absolute_import

from .CoreBu import *
from .CheckBu import *
from .FactorSellBu import *
from .FactorBuyBu import *
from .AlphaBu import *
from .BetaBu import *
from .DLBu import *
from .IndicatorBu import *
from .MLBu import *
from .MetricsBu import *
from .PickStockBu import *
from .SlippageBu import *
from .UtilBu import *
from .TLineBu import *
from .TradeBu import *
from .UmpBu import *
from .MarketBu import *
from .SimilarBu import *
from .WidgetBu import *

__all__ = ['AlphaBu', 'BetaBu', 'CheckBu', 'UmpBu', 'FactorSellBu', 'FactorSell', 'IndicatorBu', 'MarketBu', 'UtilBu',
           'SimilarBu', 'MetricsBu', 'SlippageBu', 'PickStockBu', 'CoreBu', 'TLineBu',
           'MLBu', 'DLBu', 'TradeBu', 'WidgetBu']

__version__ = '0.2.0'
__author__ = '阿布'
__weixin__ = 'abu_quant'



# -*- encoding:utf-8 -*-
"""
    选股示例因子：价格拟合角度选股因子
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np

from ..UtilBu import ABuRegUtil
from .ABuPickStockBase import AbuPickStockBase, reversed_result

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuPickRegressAngMinMax(AbuPickStockBase):
    """拟合角度选股因子示例类"""
    def _init_self(self, **kwargs):
        """通过kwargs设置拟合角度边际条件，配置因子参数"""

        # 暂时与base保持一致不使用kwargs.pop('a', default)方式
        # fit_pick中 ang > threshold_ang_min, 默认负无穷，即默认所有都符合
        self.threshold_ang_min = -np.inf
        if 'threshold_ang_min' in kwargs:
            # 设置最小角度阀值
            self.threshold_ang_min = kwargs['threshold_ang_min']

        # fit_pick中 ang < threshold_ang_max, 默认正无穷，即默认所有都符合
        self.threshold_ang_max = np.inf
        if 'threshold_ang_max' in kwargs:
            # 设置最大角度阀值
            self.threshold_ang_max = kwargs['threshold_ang_max']

    @reversed_result
    def fit_pick(self, kl_pd, target_symbol):
        """开始根据自定义拟合角度边际参数进行选股"""

        # 计算走势角度
        ang = ABuRegUtil.calc_regress_deg(kl_pd.close, show=False)
        # 根据参数进行角度条件判断
        if self.threshold_ang_min < ang < self.threshold_ang_max:
            return True
        return False

    def fit_first_choice(self, pick_worker, choice_symbols, *args, **kwargs):
        raise NotImplementedError('AbuPickRegressAng fit_first_choice unsupported now!')



# -*- encoding:utf-8 -*-
"""股票基本信息图形可视化"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging

import ipywidgets as widgets

from abc import ABCMeta, abstractmethod
from IPython.display import display

from ..CoreBu import ABuEnv
from ..CoreBu.ABuFixes import six
from ..UtilBu.ABuStrUtil import to_unicode
from ..MarketBu.ABuSymbol import search_to_symbol_dict

__author__ = '阿布'
__weixin__ = 'abu_quant'

"""基于不同系统的提示框使用partial包装title以及显示log"""
# show_msg_func = partial(show_msg, '提示', log=True)
show_msg_func = logging.info


def browser_down_csv_zip(open_browser=False):
    """浏览器打开教程使用的csv数据百度云地址"""
    try:
        if open_browser:
            import webbrowser
            webbrowser.open('https://pan.baidu.com/s/1geNZgqf', new=0, autoraise=True)
            show_msg_func(u'提取密码: gvtr')
    except:
        pass
    finally:
        logging.info(u'建议直接从百度云下载教程中使用的csv格式美股，A股，港股，币类，期货6年日k数据: ')
        logging.info(u'下载地址: https://pan.baidu.com/s/1geNZgqf')
        logging.info(u'提取密码: gvtr')
        logging.info(u'下载完成后解压zip得到\'csv\'文件夹到\'{}\'目录下'.format(ABuEnv.g_project_data_dir))


# noinspection PyUnresolvedReferences,PyProtectedMember
class WidgetBase(object):
    """界面组件基类，限定最终widget为self.widget"""

    def __call__(self):
        return self.widget

    def display(self):
        """显示使用统一display"""
        display(self.widget)


class WidgetFactorBase(six.with_metaclass(ABCMeta, WidgetBase)):
    """策略可视化基础类"""

    def __init__(self, wg_manager):
        self.wg_manager = wg_manager
        self.widget = None
        self.label_layout = widgets.Layout(width='300px', align_items='stretch')
        self.description_layout = widgets.Layout(height='150px')
        self.widget_layout = widgets.Layout(align_items='stretch', justify_content='space-between')

    @abstractmethod
    def _init_widget(self):
        """子类因子界面设置初始化"""
        pass

    @abstractmethod
    def delegate_class(self):
        """子类因子所委托的具体因子类"""
        pass


class WidgetFactorManagerBase(six.with_metaclass(ABCMeta, WidgetBase)):
    """策略管理可视化基础类"""

    def __init__(self):
        self.factor_dict = {}
        self.factor_wg_array = []
        # 策略候选池可x轴左右滚动
        self.factor_layout = widgets.Layout(overflow_x='scroll',
                                            flex_direction='row',
                                            display='flex')
        self.selected_factors = widgets.SelectMultiple(
            options=[],
            description=u'已添加的全局策略:',
            disabled=False,
            layout=widgets.Layout(width='100%', align_items='stretch')
        )
        # 已添加的全局策略可点击删除
        self.selected_factors.observe(self.remove_factor, names='value')
        # 全局策略改变通知接收序列
        self.selected_factors_obs = set()
        self.factor_box = None
        # 默认不启动可滚动因子界面，因为对外的widget版本以及os操作系统不统一
        self.scroll_factor_box = False
        # 构建具体子类的界面构建
        self._init_widget()
        if self.factor_box is None:
            raise RuntimeError('_init_widget must build factor_box!')
        self.widget = widgets.VBox([self.factor_box, self.selected_factors])

    def _sub_children(self, children, n_split):
        """将children每n_split个为一组，组装子children_group序列"""
        sub_children_cnt = int(len(children) / n_split)
        if sub_children_cnt == 0:
            sub_children_cnt = 1
        group_adjacent = lambda a, k: zip(*([iter(a)] * k))
        children_group = list(group_adjacent(children, sub_children_cnt))
        residue_ind = -(len(children) % sub_children_cnt) if sub_children_cnt > 0 else 0
        if residue_ind < 0:
            children_group.append(children[residue_ind:])
        return children_group

    def register_subscriber(self, observe):
        """注册已选策略池更新通知与BFSubscriberMixin共同作用"""
        self.selected_factors_obs.add(observe)

    def unregister_subscriber(self, observe):
        """解除注册已选策略池更新通知与BFSubscriberMixin共同作用"""
        self.selected_factors_obs.remove(observe)

    def notify_subscriber(self):
        """通知已选策略池发生改变的observe"""
        for observe in self.selected_factors_obs:
            if hasattr(observe, 'notify_subscriber'):
                observe.notify_subscriber()

    @abstractmethod
    def _init_widget(self):
        """子类因子界面设置初始化, 内部需要构建self.factor_box"""
        pass

    def refresh_factor(self):
        """已选策略池刷新，通知其它更新"""
        self.selected_factors.options = list(self.factor_dict.keys())
        self.notify_subscriber()

    def remove_factor(self, select):
        """点击从策略池中删除已选择的策略"""
        for st_key in list(select['new']):
            self.factor_dict.pop(st_key)
        self.selected_factors.options = list(self.factor_dict.keys())
        # 通知其它需要一起更新的界面进行更新
        self.notify_subscriber()

    def add_factor(self, factor_dict, factor_desc_key):
        """根据具体策略提供的策略字典对象和策略描述构建上层策略序列"""
        if factor_desc_key in self.factor_dict:
            msg = u'{} 策略已经添加过，重复添加！'.format(factor_desc_key)
            show_msg_func(msg)
            return
        self.factor_dict[factor_desc_key] = factor_dict
        self.selected_factors.options = list(self.factor_dict.keys())
        # 通知其它需要一起更新的界面进行更新
        self.notify_subscriber()


class WidgetSearchBox(WidgetBase):
    """搜索框ui界面"""

    # noinspection PyProtectedMember
    def __init__(self, search_result_callable):
        """构建股票池选股ui界面"""
        if not callable(search_result_callable):
            raise TypeError('search_result_select_func must callable!')
        # symbol搜索框构建
        self.search_bt = widgets.Button(description=u'搜索:', layout=widgets.Layout(height='10%', width='7%'))
        self.search_input = widgets.Text(
            value='',
            placeholder=u'交易代码/公司名称/拼音首字母',
            description='',
            disabled=False
        )
        self.search_input.observe(self._search_input_change, names='value')

        # symbol搜索结果框
        self.search_result = widgets.SelectMultiple(
            options=[],
            description=u'搜索结果:',
            disabled=False,
            layout=widgets.Layout(width='300px', align_items='stretch', justify_content='space-between')
        )
        self.search_result.observe(search_result_callable, names='value')
        self.search_bt.on_click(self._do_search)

        # 搜索框 ＋ 按钮 ＋ 结果框 box拼接
        sc_hb = widgets.HBox([self.search_bt, self.search_input])
        self.widget = widgets.VBox([sc_hb, self.search_result])

    # noinspection PyUnusedLocal
    def _do_search(self, bt):
        """搜索框搜索执行函数"""
        result_dict = search_to_symbol_dict(self.search_input.value)
        result_options = [u'{}:{}'.format(to_unicode(result_dict[symbol]), to_unicode(symbol))
                          for symbol in result_dict]
        self.search_result.options = result_options

    def _search_input_change(self, change):
        """当搜索输入框文字大于1个进行自动搜索"""
        search_word = change['new']
        if len(search_word) > 1:
            # 和_do_search不同这里使用fast_mode
            result_dict = search_to_symbol_dict(self.search_input.value, fast_mode=True)
            result_options = [u'{}:{}'.format(to_unicode(result_dict[symbol]), to_unicode(symbol))
                              for symbol in result_dict]
            self.search_result.options = result_options



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os

import pandas as pd

from ..CoreBu import ABuEnv
from ..UtilBu import ABuFileUtil

SIMILAR_CACHE_PATH = os.path.join(ABuEnv.g_project_cache_dir, 'similar.hdf5')


def similar_key(symbol, cmp_cnt=None, n_folds=None, start=None, end=None, corr_type=None):
    return '{}_{}_{}_{}_{}_{}'.format(symbol, cmp_cnt, n_folds, start, end, corr_type)


def dump_2_hdh5(key, obj):
    ABuFileUtil.dump_hdf5(SIMILAR_CACHE_PATH, obj, key)


def load_2_hdh5(key):
    return ABuFileUtil.load_hdf5(SIMILAR_CACHE_PATH, key)


def show_keys():
    with pd.HDFStore(SIMILAR_CACHE_PATH) as h5s:
        return h5s.keys()


def clear_cache(key=None):
    if key is not None:
        ABuFileUtil.del_hdf5(SIMILAR_CACHE_PATH, key)
    else:
        ABuFileUtil.del_file(SIMILAR_CACHE_PATH)



# -*- encoding:utf-8 -*-
"""
    基准模块，基准的作用在于交易时间范围确定，交易时间序列对齐，
    抛弃异常时间序列，交易市场范围限制，以及对比与策略的度量结果等作用
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from ..CoreBu.ABuEnv import EMarketDataSplitMode, EMarketTargetType
from ..MarketBu import ABuSymbolPd
from ..MarketBu.ABuSymbol import IndexSymbol, Symbol
from ..CoreBu import ABuEnv
from ..CoreBu.ABuBase import PickleStateMixin
from ..CoreBu.ABuFixes import six

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuBenchmark(PickleStateMixin):
    """基准类，混入PickleStateMixin，因为在abu.store_abu_result_tuple会进行对象本地序列化"""

    def __init__(self, benchmark=None, start=None, end=None, n_folds=2, rs=True, benchmark_kl_pd=None):
        if benchmark_kl_pd is not None and hasattr(benchmark_kl_pd, 'name'):
            """从金融时间序列直接构建"""
            self.benchmark = benchmark_kl_pd.name
            self.start = benchmark_kl_pd.iloc[0].date
            self.end = benchmark_kl_pd.iloc[-1].date
            self.n_folds = n_folds
            self.kl_pd = benchmark_kl_pd
            return

        if benchmark is None:
            if ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_US:
                # 美股
                benchmark = IndexSymbol.IXIC
            elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_HK:
                # 港股
                benchmark = IndexSymbol.HSI
            elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_CN:
                # a股
                benchmark = IndexSymbol.SH
            elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
                # 国内期货
                benchmark = IndexSymbol.BM_FUTURES_CN
            elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_TC:
                # 币类市场
                benchmark = IndexSymbol.TC_INX
            elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_OPTIONS_US:
                # 美股期权暂时也以IXIC做为标尺，最好是外部参数中的benchmark设置
                benchmark = IndexSymbol.IXIC
            elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL:
                # 国际期货暂时也以BM_FUTURES_GB做为标尺
                benchmark = IndexSymbol.BM_FUTURES_GB
            else:
                raise TypeError('benchmark is None AND g_market_target ERROR!')

        self.benchmark = benchmark
        self.start = start
        self.end = end
        self.n_folds = n_folds
        # 基准获取数据使用data_mode=EMarketDataSplitMode.E_DATA_SPLIT_SE，即不需要对齐其它，只需要按照时间切割
        self.kl_pd = ABuSymbolPd.make_kl_df(benchmark, data_mode=EMarketDataSplitMode.E_DATA_SPLIT_SE,
                                            n_folds=n_folds,
                                            start=start, end=end)

        if rs and self.kl_pd is None:
            # 如果基准时间序列都是none，就不要再向下运行了
            raise ValueError('CapitalClass init benchmark kl_pd is None')

    def unpick_extend_work(self, state):
        """完成 PickleStateMixin中__setstate__结束之前的工作，为kl_pd.name赋予准确的benchmark"""
        if isinstance(self.benchmark, Symbol):
            self.kl_pd.name = self.benchmark.value
        elif isinstance(self.benchmark, six.string_types):
            self.kl_pd.name = self.benchmark

    def __str__(self):
        """打印对象显示：benchmark n_folds"""
        return 'benchmark is {}, n_folds = {}'.format(self.kl_pd.name, self.n_folds)

    __repr__ = __str__

"""
    # 如果需要本地序列化很多，需要考虑存贮空间可使用LazyFunc.
    from ..UtilBu.ABuLazyUtil import LazyFunc

    class AbuBenchmark(object):
        def __init__(self, benchmark=None, start=None, end=None, n_folds=2):
            if benchmark is None:
                if ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_US:
                    benchmark = IndexSymbol.IXIC
                elif ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_HK:
                    benchmark = IndexSymbol.HSI
                else:
                    benchmark = IndexSymbol.SH
            self.benchmark = benchmark

            self.n_folds = n_folds
            self.start = start
            self.end = end

    @LazyFunc
    def kl_pd(self):
        kl_pd = ABuSymbolPd.make_kl_df(self.benchmark, data_mode=EMarketDataSplitMode.E_DATA_SPLIT_SE,
                                       n_folds=self.n_folds,
                                       start=self.start, end=self.end)

        if kl_pd is None:
            raise ValueError('CapitalClass init benchmark kl_pd is None')
        return kl_pd
"""



# -*- encoding:utf-8 -*-
"""
    示例买入择时因子
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
import pandas as pd

from .ABuFactorBuyBase import AbuFactorBuyBase, AbuFactorBuyXD, AbuFactorBuyTD, BuyCallMixin
from .ABuFactorBuyBreak import AbuFactorBuyBreak
from ..TLineBu.ABuTL import AbuTLine
from ..FactorBuyBu.ABuBuyFactorWrap import AbuLeastPolyWrap

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyAttributeOutsideInit
class AbuSDBreak(AbuFactorBuyXD, BuyCallMixin):
    """示例买入因子： 在AbuFactorBuyBreak基础上进行降低交易频率，提高系统的稳定性处理"""

    def _init_self(self, **kwargs):
        """
        :param kwargs: kwargs可选参数poly值，poly在fit_month中和每一个月大盘计算的poly比较，
        若是大盘的poly大于poly认为走势震荡，poly默认为2
        """
        super(AbuSDBreak, self)._init_self(**kwargs)
        # poly阀值，self.poly在fit_month中和每一个月大盘计算的poly比较，若是大盘的poly大于poly认为走势震荡
        self.poly = kwargs.pop('poly', 2)
        # 是否封锁买入策略进行择时交易
        self.lock = False

    def fit_month(self, today):
        # fit_month即在回测策略中每一个月执行一次的方法
        # 策略中拥有self.benchmark，即交易基准对象，AbuBenchmark实例对象，benchmark.kl_pd即对应的市场大盘走势
        benchmark_df = self.benchmark.kl_pd
        # 拿出大盘的今天
        benchmark_today = benchmark_df[benchmark_df.date == today.date]
        if benchmark_today.empty:
            return 0
        # 要拿大盘最近一个月的走势，准备切片的start，end
        end_key = int(benchmark_today.iloc[0].key)
        start_key = end_key - 20
        if start_key < 0:
            return False

        # 使用切片切出从今天开始向前20天的数据
        benchmark_month = benchmark_df[start_key:end_key + 1]
        # 通过大盘最近一个月的收盘价格做为参数构造AbuTLine对象
        benchmark_month_line = AbuTLine(benchmark_month.close, 'benchmark month line')
        # 计算这个月最少需要几次拟合才能代表走势曲线
        least = benchmark_month_line.show_least_valid_poly(show=False)

        if least >= self.poly:
            # 如果最少的拟合次数大于阀值self.poly，说明走势成立，大盘非震荡走势，解锁交易
            self.lock = False
        else:
            # 如果最少的拟合次数小于阀值self.poly，说明大盘处于震荡走势，封锁策略进行交易
            self.lock = True

    def fit_day(self, today):
        if self.lock:
            # 如果封锁策略进行交易的情况下，策略不进行择时
            return None

        # 今天的收盘价格达到xd天内最高价格则符合买入条件
        if today.close == self.xd_kl.close.max():
            return self.buy_tomorrow()


@AbuLeastPolyWrap()
class AbuTwoDayBuy(AbuFactorBuyTD, BuyCallMixin):
    """示例AbuLeastPolyWrap，混入BuyCallMixin，即向上突破触发买入event"""

    def _init_self(self, **kwargs):
        """简单示例什么都不编写了"""
        pass

    def fit_day(self, today):
        """
        针对每一个交易日拟合买入交易策略，今天涨，昨天涨就买
        :param today: 当前驱动的交易日金融时间序列数据
        :return:
        """
        # 今天的涨幅
        td_change = today.p_change
        # 昨天的涨幅
        yd_change = self.yesterday.p_change

        if td_change > 0 and 0 < yd_change < td_change:
            # 连续涨两天, 且今天的涨幅比昨天还高 －>买入, 用到了今天的涨幅，只能明天买
            return self.buy_tomorrow()
        return None


class AbuWeekMonthBuy(AbuFactorBuyBase, BuyCallMixin):
    """策略示例每周买入一次或者每一个月买入一次"""

    def _init_self(self, **kwargs):
        """kwargs可选参数：is_buy_month，bool默认True一个月买入一次, False一周买入一次"""
        self.is_buy_month = kwargs.pop('is_buy_month', True)

    def fit_day(self, today):
        """
        :param today: 当前驱动的交易日金融时间序列数据
        """
        if self.is_buy_month and today.exec_month or not self.is_buy_month and today.exec_week:
            # 没有用到今天的任何数据，直接今天买入
            return self.buy_today()


class AbuFactorBuyBreakUmpDemo(AbuFactorBuyBreak):
    """示例组织裁判进行更复杂的综合裁决 扩展AbuFactorBuyBreak组织裁判进行更复杂的综合裁决"""

    def make_ump_block_decision(self, ml_feature_dict):
        """
        进行裁判之间的拦截配合, 简单示例，只要加起来大于2个就算配合成功，拦截
        :param ml_feature_dict: 需要决策的当前买入时刻交易特征dict
        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """

        ump = self.ump_manger
        # 统计角度主裁对应这次交易命中的分类簇个数
        deg_hit_cnt = ump.ump_main_deg.predict_hit_kwargs(**ml_feature_dict)
        # 统计跳空主裁对应这次交易命中的分类簇个数
        jump_hit_cnt = ump.ump_main_jump.predict_hit_kwargs(**ml_feature_dict)
        # 统计波动主裁对应这次交易命中的分类簇个数
        wave_hit_cnt = ump.ump_main_wave.predict_hit_kwargs(**ml_feature_dict)
        # 统计价格主裁对应这次交易命中的分类簇个数
        price_hit_cnt = ump.ump_main_price.predict_hit_kwargs(**ml_feature_dict)
        # 进行裁判之间的拦截配合, 简单示例，只要加起来大于2个就算配合成功，拦截
        if deg_hit_cnt + jump_hit_cnt + wave_hit_cnt + price_hit_cnt > 2:
            return True
        return False


class AbuFactorBuyBreakReocrdHitDemo(AbuFactorBuyBreak):
    """示例让裁判自己学习怎么配合，自己做出最正确的判断"""

    def make_ump_block_decision(self, ml_feature_dict):
        """
        即是可以再次根据裁判之间的配合数据进行训练学习，让裁判自己学习怎么配合，自己做出最正确的判断，
        而不是像上面的示例使用固定值3来做为裁决阀值，AbuFactorBuyBreakReocrdHitDemo类似
        AbuFactorBuyBreakUmpDemo但是不对交易进行决策，只是把每一个裁判的对应交易命中的分类簇个数进行记录，更新在特征数据里
        :param ml_feature_dict: 需要决策的当前买入时刻交易特征dict
        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """
        ump = self.ump_manger
        # 统计角度主裁对应这次交易命中的分类簇个数
        deg_hit_cnt = ump.ump_main_deg.predict_hit_kwargs(**ml_feature_dict)
        # 统计跳空主裁对应这次交易命中的分类簇个数
        jump_hit_cnt = ump.ump_main_jump.predict_hit_kwargs(**ml_feature_dict)
        # 统计波动主裁对应这次交易命中的分类簇个数
        wave_hit_cnt = ump.ump_main_wave.predict_hit_kwargs(**ml_feature_dict)
        # 统计价格主裁对应这次交易命中的分类簇个数
        price_hit_cnt = ump.ump_main_price.predict_hit_kwargs(**ml_feature_dict)

        ml_feature_dict.update({'deg_hit_cnt': deg_hit_cnt, 'jump_hit_cnt': jump_hit_cnt,
                                'wave_hit_cnt': wave_hit_cnt, 'price_hit_cnt': price_hit_cnt})

        return False


class AbuFactorBuyBreakHitPredictDemo(AbuFactorBuyBreak):
    """
        继续继承AbuFactorBuyBreak复写make_ump_block_decision，
        区别是使用AbuFactorBuyBreakReocrdHitDemo的学习成果hit_ml
        对几个裁判这次交易命中的分类簇个数组成矢量特征进行predict，
        拦截预测结果为-1的交易
    """

    def _init_self(self, **kwargs):
        """
            与AbuFactorBuyBreak基本相同，唯一区别是关键子参数中添加了通过AbuFactorBuyBreakUmpDemo记录训练好的决策器
            self.hit_ml = kwargs['hit_ml']
        """
        super(AbuFactorBuyBreakHitPredictDemo, self)._init_self(**kwargs)
        # 添加了通过AbuFactorBuyBreakUmpDemo记录训练好的决策器
        self.hit_ml = kwargs['hit_ml']

    def make_ump_block_decision(self, ml_feature_dict):
        """
        用回测的数据进行训练后再次反过来指导回测，结果是没有意义的，
        这里的示例只是为了容易理解什么叫做：让裁判自己学习怎么配合，
        自己做出最正确的判断，更详细完整的示例会在之后的章节中示例讲解，
        请关注公众号的更新提醒
        :param ml_feature_dict: 需要决策的当前买入时刻交易特征dict
        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """
        ump = self.ump_manger
        # 统计角度主裁对应这次交易命中的分类簇个数
        deg_hit_cnt = ump.ump_main_deg.predict_hit_kwargs(**ml_feature_dict)
        # 统计跳空主裁对应这次交易命中的分类簇个数
        jump_hit_cnt = ump.ump_main_jump.predict_hit_kwargs(**ml_feature_dict)
        # 统计波动主裁对应这次交易命中的分类簇个数
        wave_hit_cnt = ump.ump_main_wave.predict_hit_kwargs(**ml_feature_dict)
        # 统计价格主裁对应这次交易命中的分类簇个数
        price_hit_cnt = ump.ump_main_price.predict_hit_kwargs(**ml_feature_dict)

        result = self.hit_ml.predict([deg_hit_cnt, jump_hit_cnt, wave_hit_cnt, price_hit_cnt])[0]
        if result == -1:
            return True
        return False


class AbuBTCDayBuy(AbuFactorBuyBase, BuyCallMixin):
    """
        比特币日交易策略：

        1. 买入条件1: 当日这100个股票60%以上都是上涨的
        2. 买入条件2: 使用在第12节：机器学习与比特币示例中编写的：信号发出今天比特币会有大行情
    """

    def _init_self(self, **kwargs):
        from ..MarketBu import ABuSymbolPd

        # 市场中与btc最相关的top个股票
        self.btc_similar_top = kwargs.pop('btc_similar_top')
        # 超过多少个相关股票今天趋势相同就买入
        self.btc_vote_val = kwargs.pop('btc_vote_val', 0.60)

        def _collect_kl(sim_line):
            """在初始化中将所有相关股票的对应时间的k线数据进行收集"""
            start = self.kl_pd.iloc[0].date
            end = self.kl_pd.iloc[-1].date
            kl = ABuSymbolPd.make_kl_df(sim_line.symbol, start=start, end=end)
            self.kl_dict[sim_line.symbol] = kl

        self.kl_dict = {}
        # k线数据进行收集到类字典对象self.kl_dict中
        self.btc_similar_top.apply(_collect_kl, axis=1)

    # noinspection PyUnresolvedReferences
    def fit_day(self, today):
        """
        :param today: 当前驱动的交易日金融时间序列数据
        """
        # 忽略不符合买入的天（统计周期内前两天, 因为btc的机器学习特证需要三天交易数据）
        if self.today_ind < 2:
            return None

        # 今天，昨天，前天三天的交易数据进行特证转换
        btc = self.kl_pd[self.today_ind - 2:self.today_ind + 1]
        # 三天的交易数据进行转换后得到btc_today_x
        btc_today_x = self.make_btc_today(btc)

        # btc_ml并没有在这里传入，实际如果要使用，需要对外部的btc_ml进行本地序列化后，构造读取本地
        # 买入条件2: 使用在第12节：机器学习与比特币示例中编写的：信号发出今天比特币会有大行情
        if btc_ml.predict(btc_today_x):
            # 买入条件1: 当日这100个股票60%以上都是上涨的
            vote_val = self.similar_predict(today.date)
            if vote_val > self.btc_vote_val:
                # 没有使用当天交易日的close等数据，且btc_ml判断的大波动是当日，所以当日买入
                return self.buy_today()

    # noinspection PyUnresolvedReferences
    def make_btc_today(self, sib_btc):
        """构造比特币三天数据特证"""
        from ..UtilBu import ABuScalerUtil

        sib_btc['big_wave'] = (sib_btc.high - sib_btc.low) / sib_btc.pre_close > 0.55
        sib_btc['big_wave'] = sib_btc['big_wave'].astype(int)
        sib_btc_scale = ABuScalerUtil.scaler_std(
            sib_btc.filter(['open', 'close', 'high', 'low', 'volume', 'pre_close',
                            'ma5', 'ma10', 'ma21', 'ma60', 'atr21', 'atr14']))
        # 把标准化后的和big_wave，date_week连接起来
        sib_btc_scale = pd.concat([sib_btc['big_wave'], sib_btc_scale, sib_btc['date_week']], axis=1)

        # 抽取第一天，第二天的大多数特征分别改名字以one，two为特征前缀，如：one_open，one_close，two_ma5，two_high.....
        a0 = sib_btc_scale.iloc[0].filter(['open', 'close', 'high', 'low', 'volume', 'pre_close',
                                           'ma5', 'ma10', 'ma21', 'ma60', 'atr21', 'atr14', 'date_week'])
        a0.rename(index={'open': 'one_open', 'close': 'one_close', 'high': 'one_high', 'low': 'one_low',
                         'volume': 'one_volume', 'pre_close': 'one_pre_close',
                         'ma5': 'one_ma5', 'ma10': 'one_ma10', 'ma21': 'one_ma21',
                         'ma60': 'one_ma60', 'atr21': 'one_atr21', 'atr14': 'one_atr14',
                         'date_week': 'one_date_week'}, inplace=True)

        a1 = sib_btc_scale.iloc[1].filter(['open', 'close', 'high', 'low', 'volume', 'pre_close',
                                           'ma5', 'ma10', 'ma21', 'ma60', 'atr21', 'atr14', 'date_week'])
        a1.rename(index={'open': 'two_open', 'close': 'two_close', 'high': 'two_high', 'low': 'two_low',
                         'volume': 'two_volume', 'pre_close': 'two_pre_close',
                         'ma5': 'two_ma5', 'ma10': 'two_ma10', 'ma21': 'two_ma21',
                         'ma60': 'two_ma60', 'atr21': 'two_atr21', 'atr14': 'two_atr14',
                         'date_week': 'two_date_week'}, inplace=True)
        # 第三天的特征只使用'open', 'low', 'pre_close', 'date_week'，该名前缀today，如today_open，today_date_week
        a2 = sib_btc_scale.iloc[2].filter(['big_wave', 'open', 'low', 'pre_close', 'date_week'])
        a2.rename(index={'open': 'today_open', 'low': 'today_low',
                         'pre_close': 'today_pre_close',
                         'date_week': 'today_date_week'}, inplace=True)
        # 将抽取改名字后的特征连接起来组合成为一条新数据，即3天的交易数据特征－>1条新的数据
        btc_today = pd.DataFrame(pd.concat([a0, a1, a2], axis=0)).T

        # 开始将周几进行离散处理
        dummies_week_col = btc_ml.df.filter(regex='(^one_date_week_|^two_date_week_|^today_date_week_)').columns
        dummies_week_df = pd.DataFrame(np.zeros((1, len(dummies_week_col))), columns=dummies_week_col)

        # 手动修改每一天的one hot
        one_day_key = 'one_date_week_{}'.format(btc_today.one_date_week.values[0])
        dummies_week_df[one_day_key] = 1
        two_day_key = 'two_date_week_{}'.format(btc_today.two_date_week.values[0])
        dummies_week_df[two_day_key] = 1
        today_day_key = 'today_date_week_{}'.format(btc_today.today_date_week.values[0])
        dummies_week_df[today_day_key] = 1
        btc_today.drop(['one_date_week', 'two_date_week', 'today_date_week'], inplace=True, axis=1)
        btc_today = pd.concat([btc_today, dummies_week_df], axis=1)
        return btc_today.as_matrix()[:, 1:]

    def similar_predict(self, today_date):
        """与比特币在市场中最相关的top100个股票已各自今天的涨跌结果进行投票"""

        def _predict_vote(sim_line, _today_date):
            kl = self.kl_dict[sim_line.symbol]
            if kl is None:
                return -1 * sim_line.vote_direction > 0
            kl_today = kl[kl.date == _today_date]
            if kl_today is None or kl_today.empty:
                return -1 * sim_line.vote_direction > 0
            # 需要 * sim_line.vote_direction，因为负相关的存在
            return kl_today.p_change.values[0] * sim_line.vote_direction > 0

        vote_result = self.btc_similar_top.apply(_predict_vote, axis=1, args={today_date, })
        # 将所有投票结果进行统计，得到与比特币最相关的这top100个股票的今天投票结果
        vote_val = 1 - vote_result.value_counts()[False] / vote_result.value_counts().sum()
        return vote_val



# -*- encoding:utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import pandas as pd

import math
import logging

from os import path

from abupy.CrawlBu import ABuXqFile
from ..CoreBu.ABuEnvProcess import add_process_env_sig
from ..CrawlBu.ABuXqFile import map_stock_list
from .ABuXqCrawlImp import StockInfoListBrower, StockListCrawlBrower
from .ABuXqCrawlImp import NavHQCrawlBrower

from ..CoreBu.ABuEnv import g_project_rom_data_dir
from ..CoreBu.ABuParallel import Parallel, delayed
from ..UtilBu import ABuDTUtil

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import map, reduce, filter

__author__ = '小青蛙'
__weixin__ = 'abu_quant'

_NAV_PATH = path.join(g_project_rom_data_dir, 'hq_nav.txt')


@add_process_env_sig
def __crawl_stock_parallel(market, url):
    with StockListCrawlBrower(url) as crawler:
        name, symbol = crawler.crawl()
        df = pd.DataFrame({'name': name, 'symbol': symbol})
        if market == 'CN':
            # 确保都是A股正确的symbol
            df = df.loc[
                list(
                    map(lambda s: (s[:2] in ['SZ', 'SH', 'sz', 'sh']) and str(s[2:]).isdigit(), df['symbol'].tolist()))]
            df['market'] = df['symbol'].map(lambda s: s[:2])
        else:
            df['market'] = market

        df.to_csv(map_stock_list(market), index=False)


@add_process_env_sig
def __crawl_stock_info_parallel(market, symbols, process):
    with StockInfoListBrower(market, symbols) as stockInfo:
        rest = stockInfo.crawl(process=process)
        print(rest)


def hq_nav():
    """
    :return: dict
    一级菜单：类别，如。美股，港股，沪深，基金，债券等
    二级菜单：如，行业分类，美股一览等
    三级菜单：可能会有(,比如行业就需细分)
    """
    with NavHQCrawlBrower() as nav:
        return nav.crawl()


@ABuDTUtil.warnings_filter
def crawl_stock_code(markets=('CN', 'HK', 'US')):
    """
    从雪球获取市场股市代码
    :param markets: 市场类型
    """
    nav = hq_nav()
    urls = dict()
    for market_nav in nav:
        for first_menu in nav[market_nav]:
            if first_menu in [u'沪深一览'] and 'CN' in markets:
                urls['CN'] = nav[market_nav][first_menu]
            if first_menu in [u'港股一览'] and 'HK' in markets:
                urls['HK'] = nav[market_nav][first_menu]
            if first_menu in [u'美股一览'] and 'US' in markets:
                urls['US'] = nav[market_nav][first_menu]

    Parallel(n_jobs=-1)(delayed(__crawl_stock_parallel)(m, urls[m]) for m in urls)


def crawl_stock_info(markets):
    """
    获取股票信息，例如，股市简介，行业，市值等信息
    :param markets:
    """
    _markets = []
    _parts = []
    for m in markets:
        symbols = ABuXqFile.read_stock_symbol_list(m)
        interval = 1000

        # 将 symbols 划分成一组1000个
        for s in range(int(math.ceil(len(symbols) / interval))):
            st = s * interval
            en = (s + 1) * interval if (s + 1) * interval <= len(symbols) else len(symbols)
            _parts.append(symbols[st:en])
            _markets.append(m)
    # 开启len(_parts)个进程同时
    print('任务数{}'.format(len(_parts)))

    parallel = Parallel(n_jobs=-1)
    parallel(delayed(__crawl_stock_info_parallel)(m, p, i)
             for i, (m, p) in enumerate(zip(_markets, _parts)))


def ensure_symbol(symbol):
    """
    保证本地的股市代码库纯在该symbol才能查看stock info
    A 股 sh、sz开头 + 6位数字
    HK 股 5位数字
    UK 股 英文代码
    :param symbol:
    :return:
    """
    symbol = str(symbol).upper()
    if (len(symbol) == 6 and str(symbol).isdigit()) or (
                        len(symbol) == 8 and symbol[:2] in ['SZ', 'SH'] and str(symbol[2:]).isdigit()):
        market = 'CN'
    elif (len(symbol) == 5 and str(symbol).isdigit()) or (
                        len(symbol) == 7 and symbol[:2] in ['HK'] and str(symbol[2:]).isdigit()):
        market = 'HK'
    else:
        market = 'US'
    stock_list = ABuXqFile.read_stock_symbol_list(market)

    if market in ['CN']:
        # symbol自带 sh或sz
        if len(symbol) == 8:
            return market, symbol
        # 此刻 symbol是6位数字
        sh_code = 'SH{}'.format(symbol)
        sz_code = 'SZ{}'.format(symbol)
        is_sh = sh_code in stock_list
        is_sz = sz_code in stock_list

        if is_sz and is_sh:
            # 即使 上证又是深圳股
            sh_name = ABuXqFile.read_stock_symbol_name(market, sh_code)
            sz_name = ABuXqFile.read_stock_symbol_name(market, sz_code)
            logging.error(
                '找到两个关于{}的股票{}({}),{}({}),请改写{}成{}或{}'.format(
                    symbol, sh_code, sh_name, sz_code, sz_name, symbol, sz_code, sh_code))
            return market, None
        elif is_sh:
            return market, sh_code
        elif is_sz:
            return sz_code
        else:
            pass
    elif market in ['US', 'HK']:
        if symbol in stock_list:
            return market, symbol

    logging.error('没有找到{}相关的股票'.format(symbol))
    return market, None


def update_all(markets=('US', 'CN', 'HK')):
    crawl_stock_code(markets)
    crawl_stock_info(markets)
    ABuXqFile.merge_stock_info_to_stock_list(markets)
    ABuXqFile.fix_xq_columns_name()


def query_symbol_info(symbol):
    m, symbol = ensure_symbol(symbol)
    return None if symbol is None else ABuXqFile.query_a_stock(m, symbol)



from __future__ import absolute_import

from .ABuPickBase import AbuPickTimeWorkBase, AbuPickStockWorkBase

from .ABuPickStockMaster import AbuPickStockMaster
from .ABuPickStockWorker import AbuPickStockWorker

from .ABuPickTimeWorker import AbuPickTimeWorker
from .ABuPickTimeMaster import AbuPickTimeMaster

from . import ABuPickStockExecute
from . import ABuPickTimeExecute
# noinspection all
from . import ABuAlpha as alpha

__all__ = [
    'AbuPickTimeWorkBase',
    'AbuPickStockWorkBase',
    'AbuPickStockMaster',
    'AbuPickStockWorker',
    'AbuPickTimeWorker',
    'AbuPickTimeMaster',

    'ABuPickStockExecute',
    'ABuPickTimeExecute',
    'alpha'
]



# -*- encoding:utf-8 -*-
"""直观可视化制作qcut的bins点"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import math
import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import itertools

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import filter

__all__ = ['show_orders_hist']


def show_orders_hist(df, feature_columns, show=True, only_hist=True, show_pie=False):
    """
    可视化统计feature_columns序列所指定的特征在df中的直方图或者饼状图，
    根据参数only_hist是否进行pd.qcut统计输出

    eg:
        from abupy import AbuML, ml
        ttn_raw = AbuML.load_ttn_raw_df()
        ml.show_orders_hist(ttn_raw, ['Age', 'Fare', 'Pclass'])

    :param df: pd.DataFrame对象
    :param feature_columns: 特征名称序列，eg：['Age', 'Fare', 'Pclass']
    :param show: 是否可视化直方图或者饼状图
    :param show_pie: 是否优先考虑绘制饼状图，默认false
    :param only_hist: 是否进行pd.qcut统计输出
    """
    if not isinstance(df, pd.DataFrame):
        logging.info('df must pd.DataFrame, not type {}'.format(type(df)))
        return

    # 第一步过滤不在在特征列中的feature_columns元素
    feature_columns = list(filter(lambda x: df.columns.tolist().count(x) > 0, feature_columns))
    # 第二步过滤feature_columns元素中类型不是int或者float的
    feature_columns = list(
        filter(
            lambda x: df[x].dtype == int or df[x].dtype == float or df[x].dtype == np.uint or df[x].dtype == np.uint8,
            feature_columns))
    # 第三步过滤feature_columns元素中所指特征列中unique==1的，eg：1列全是1，全是0，没办法做bin
    feature_columns = list(filter(lambda x: len(np.unique(df[x])) > 1, feature_columns))

    axs_list = None
    if len(feature_columns) == 0:
        # 晒没了的情况，直接返回
        logging.info('{}\n{}\nnot exist! or unique==1!, or dtype != int or float'.format(
            df.columns, df.dtypes))
        return

    if show:
        # 如果可视化直方图，先确定子画布列数，一行放两个，取math.ceil，eg：3 ／2 ＝ 2
        n_rows = int(math.ceil(len(feature_columns) / 2))
        # 行高取5，总高度：n_rows * 5
        fig_h = n_rows * 5
        # plt.subplots生成子画布
        _, axs = plt.subplots(nrows=n_rows, ncols=2, figsize=(14, fig_h))
        # 如果是多于1个的即展开字画本序列为1d序列
        axs_list = axs if n_rows == 1 else list(itertools.chain.from_iterable(axs))

    for ind, feature in enumerate(feature_columns):
        feature_unique = len(np.unique(df[feature]))
        ax = None
        if axs_list is not None:
            ax = axs_list[ind]
            ax.set_title(feature)
        if show_pie and feature_unique < 10:
            # 如果特征的值unique < 10个，通过value_counts直接画饼图
            df[feature].value_counts().plot(ax=ax, kind='pie')
        else:
            # 画直方图
            bins = int(feature_unique / 50) if feature_unique / 50 > 10 else 10
            df[feature].hist(ax=ax, bins=bins)

        if only_hist:
            # 只做可视化就continue
            continue

        try:
            # qcut切分10等份
            cats = pd.qcut(df[feature], 10)
        except Exception:
            # 某一个数据超出q的数量导致无法分
            import pandas.core.algorithms as algos
            bins = algos.quantile(np.unique(df[feature]), np.linspace(0, 1, 10 + 1))
            # noinspection PyProtectedMember,PyUnresolvedReferences
            cats = pd.tools.tile._bins_to_cuts(df[feature], bins, include_lowest=True)

        logging.info('{0} show hist and qcuts'.format(feature))
        """
            Age show hist and qcuts
            (31.8, 36]    91
            (14, 19]      87
            (41, 50]      78
            [0.42, 14]    77
            (22, 25]      70
            (19, 22]      67
            (28, 31.8]    66
            (50, 80]      64
            (25, 28]      61
            (36, 41]      53
            Name: Age, dtype: int64
        """
        logging.info(cats.value_counts())



from __future__ import absolute_import

from .ABuPositionBase import AbuPositionBase
from .ABuAtrPosition import AbuAtrPosition
from .ABuKellyPosition import AbuKellyPosition
# noinspection all
from . import ABuBeta as beta

__all__ = [
    'AbuPositionBase',
    'AbuAtrPosition',
    'AbuKellyPosition',
    'beta'
]



# -*- encoding:utf-8 -*-
"""封装AbuML为业务逻辑层进行规范模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from abc import ABCMeta, abstractmethod

import pandas as pd
import numpy as np

from .ABuML import AbuML
from ..CoreBu.ABuFixes import six
from ..CoreBu import ABuEnv
from ..MarketBu import ABuSymbolPd
from ..IndicatorBu import ABuNDMa
from ..UtilBu import ABuScalerUtil

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuMLPd(six.with_metaclass(ABCMeta, object)):
    """封装AbuML的上层具体业务逻辑类"""

    def __init__(self, **kwarg):
        """
        从kwarg中输入数据或者，make_xy中本身生产数据，在做完
        make_xy之后，类变量中一定要有x，y和df，使用AbuML继续
        构造self.fiter
        :param kwarg: 直接透传给make_xy的关键子参数，没有必须的参数
        """
        self.make_xy(**kwarg)
        if not hasattr(self, 'x') or not hasattr(self, 'y') \
                or not hasattr(self, 'df'):
            raise ValueError('make_xy failed! x, y not exist!')
        # noinspection PyUnresolvedReferences
        self.fiter = AbuML(self.x, self.y, self.df)

    @abstractmethod
    def make_xy(self, **kwarg):
        """
        子类需要完成的abstractmethod方法，可以从**kwarg中得到数据
        或者make_xy中本身生产数据，但在make_xy之后，类变量中一定要有
        x，y和df
        """
        pass

    def __getattr__(self, item):
        """
        使用ABuML对象self.fiter做为方法代理:
            return getattr(self.fiter, item)
        即AbuMLPd中可以使用ABuML类对象中任何方法
        """
        return getattr(self.fiter, item)

    def __call__(self):
        """
        方便外面直接call，不用每次去get
        :return: self.fiter
        """
        return self.fiter


# noinspection PyAttributeOutsideInit
class ClosePredict(AbuMLPd):
    """
        示例AbuMLPd基本使用:

        获取usTSLA的沙盒测试数据，将收盘价格做为y，
        开盘，最高，最低，昨收，周几组成x矩阵，通过
        训练，预测收盘价格
    """

    def make_xy(self, **kwarg):
        """
            make_xy中读取usTSLA金融时间序列数据，使用'open', 'high', 'low', 'pre_close', 'date_week'
            做为特征列x，close即收盘价格为y，更多AbuMLPd使用阅读AbuUmpMainDeg等ump类实行
        """

        # 从沙盒中读取测试数据
        ABuEnv.enable_example_env_ipython()
        tsla = ABuSymbolPd.make_kl_df('usTSLA')
        ABuEnv.disable_example_env_ipython()

        # 留五个做为测试，其它的都做训练
        train_df = tsla[:-5]
        # make_xy中需要确定self.df
        self.df = train_df.filter(['close', 'open', 'high', 'low', 'pre_close', 'date_week'])
        tsla_matrix = self.df.as_matrix()
        # close列做为y，make_xy中需要确定self.y
        self.y = tsla_matrix[:, 0]
        # 'open', 'high', 'low', 'pre_close', 'date_week'做为x
        self.x = tsla_matrix[:, 1:]

        # 最后5个交易日做为测试数据, 只做为AbuMLPd使用示例
        test_df = tsla[-5:]
        tsla_matrix = test_df.filter(['close', 'open', 'high', 'low', 'pre_close', 'date_week']).as_matrix()
        self.y_test = tsla_matrix[:, 0]
        self.x_test = tsla_matrix[:, 1:]


def test_close_predict():
    """
        示例通过ClosePredict以及AbuMLPd的使用:
        eg:
            from abupy.MLBu.ABuMLPd import test_close_predict
            test_close_predict()
    """

    close_predict = ClosePredict()
    # ClosePredict中的数据为连续数据，属于回归，ABuML中自动会使用回归器，使用adaboost_regressor_best
    close_predict.adaboost_regressor_best()
    """
        AdaBoostRegressor(base_estimator=DecisionTreeRegressor(criterion='mse', max_depth=None, max_features=None,
           max_leaf_nodes=None, min_impurity_split=1e-07,
           min_samples_leaf=1, min_samples_split=2,
           min_weight_fraction_leaf=0.0, presort=False, random_state=None,
           splitter='best'),
         learning_rate=1.0, loss='linear', n_estimators=450,
         random_state=None)
    """
    # 可以使用ABuML中的所有方法，因为AbuMLPd使用__getattr__做了方法代理
    close_predict.plot_decision_function()
    close_predict.plot_graphviz_tree()
    # 训练数据
    close_predict.fit()

    for test_ind in np.arange(0, 5):
        predict = close_predict.predict(close_predict.x_test[test_ind])
        print('predict close is {:.3f}, actual close is {:.3f}'.format(predict[0], close_predict.y_test[test_ind]))
        """
            predict close is 228.100, actual close is 228.360
            predict close is 223.430, actual close is 220.500
            predict close is 220.690, actual close is 222.270
            predict close is 228.920, actual close is 230.010
            predict close is 228.280, actual close is 225.930
        """

    close_predict.feature_selection()
    """
        eg:
            RFE selection
                   ranking support
        open             2   False
        high             1    True
        low              1    True
        pre_close        3   False
        date_week        4   False
    """


# noinspection PyAttributeOutsideInit,PyUnresolvedReferences,PyTypeChecker
class BtcBigWaveClf(AbuMLPd):
    """
        任何大的决策其实都是由很多看极起来极不起眼的小事组成的，如果我们是做比特币日内的交易者，首先你需要判断今天适不适合做交易，
        做出这个判断的依据里有一条即是今天的波动需要足够大
    """

    def __init__(self, **kwarg):
        """
            如果传递了btc数据，说明不是运行：
            12-机器学习与比特币示例(ABU量化使用文档)
        """
        self.btc = kwarg.pop('btc', None)
        super(BtcBigWaveClf, self).__init__(**kwarg)

    def make_xy(self, **kwarg):
        if self.btc is None:
            # 从沙盒中读取测试数据
            ABuEnv.enable_example_env_ipython()
            btc = ABuSymbolPd.make_kl_df('btc', start='2013-09-01', end='2017-07-26')
            ABuEnv.disable_example_env_ipython()
        else:
            btc = self.btc
        # .055的日震荡幅度可以成做大波动的交易对比特币来说，下面对数据添加新列big_wave
        btc['big_wave'] = (btc.high - btc.low) / btc.pre_close > 0.055
        btc['big_wave'] = btc['big_wave'].astype(int)

        if self.btc is None:
            # 如果是12-机器学习与比特币示例(ABU量化使用文档)，保留60天数据
            # 首先切割训练集和测试集，保留最后60天走势数据做为测试集数据
            btc_train_raw = btc[:-60]
            btc_test_raw = btc[-60:]
        else:
            btc_train_raw = btc
            btc_test_raw = None

        # 下面为训练集和测试集数据都加上5，10，21，60日均线特征
        def calc_ma(tc, p_ma):
            ma_key = 'p_ma{}'.format(p_ma)
            tc[ma_key] = ABuNDMa.calc_ma_from_prices(tc.close, p_ma, min_periods=1)

        for ma in [5, 10, 21, 60]:
            calc_ma(btc_train_raw, ma)
            if btc_test_raw is not None:
                calc_ma(btc_test_raw, ma)
        # 下面使用训练集数据btc_train_raw做为参数抽取组合特征，重新组合好的特征
        btc_train0 = self.btc_siblings_df(btc_train_raw)
        # 由于每3条连续交易日数据组合成一个特征，只要向前跳一条数据进行特征组合抽取即可以得到另一组新特征
        btc_train1 = self.btc_siblings_df(btc_train_raw[1:])
        btc_train2 = self.btc_siblings_df(btc_train_raw[2:])

        # 把周几这个特征使用pd.get_dummies进行离散化处理，使得所有特征值的范围都在0-1之间
        btc_train = pd.concat([btc_train0, btc_train1, btc_train2])
        btc_train.index = np.arange(0, btc_train.shape[0])
        dummies_one_week = pd.get_dummies(btc_train['one_date_week'], prefix='one_date_week')
        dummies_two_week = pd.get_dummies(btc_train['two_date_week'], prefix='two_date_week')
        dummies_today_week = pd.get_dummies(btc_train['today_date_week'], prefix='today_date_week')
        btc_train.drop(['one_date_week', 'two_date_week', 'today_date_week'], inplace=True, axis=1)
        # make_xy中需要确定self.df
        self.df = pd.concat([btc_train, dummies_one_week, dummies_two_week, dummies_today_week], axis=1)
        # make_xy中需要确定x, y
        train_matrix = self.df.as_matrix()
        self.y = train_matrix[:, 0]
        self.x = train_matrix[:, 1:]

        if btc_test_raw is not None:
            # 下面将前面保留切割的60条测试数据进行特征抽取组合，方式和抽取训练集时一样
            btc_test0 = self.btc_siblings_df(btc_test_raw)
            btc_test1 = self.btc_siblings_df(btc_test_raw[1:])
            btc_test2 = self.btc_siblings_df(btc_test_raw[2:])
            btc_test = pd.concat([btc_test0, btc_test1, btc_test2])
            btc_test.index = np.arange(0, btc_test.shape[0])
            dummies_one_week = pd.get_dummies(btc_test['one_date_week'], prefix='one_date_week')
            dummies_two_week = pd.get_dummies(btc_test['two_date_week'], prefix='two_date_week')
            dummies_today_week = pd.get_dummies(btc_test['today_date_week'], prefix='today_date_week')
            btc_test.drop(['one_date_week', 'two_date_week', 'today_date_week'], inplace=True, axis=1)
            self.btc_test = pd.concat([btc_test, dummies_one_week, dummies_two_week, dummies_today_week], axis=1)
            # 测试集数据构建
            matrix_test = self.btc_test.as_matrix()
            self.y_test = matrix_test[:, 0]
            self.x_test = matrix_test[:, 1:]

    # noinspection PyMethodMayBeStatic
    def btc_siblings_df(self, btc_raw):
        """
        * 首先将所有交易日以3个为一组，切割成多个子df，即每一个子df中有3个交易日的交易数据
        * 使用数据标准化将连续3天交易日中的连续数值特征进行标准化操作
        * 抽取第一天，第二天的大多数特征分别改名字以one，two为特征前缀，如：one_open，one_close，two_ma5，two_high.....,
        * 第三天的特征只使用'open', 'low', 'pre_close', 'date_week'，该名前缀today，如today_open，today_date_week
        * 第三天的抽取了'big_wave'，其将在之后做为y
        * 将抽取改名字后的特征连接起来组合成为一条新数据，即3天的交易数据特征－>1条新的数据

        :param btc_raw: btc走势数据，pd.DataFrame对象
        :return: 重新组合好的特征数据，pd.DataFrame对象
        """

        # 将所有交易日以3个为一组，切割成多个子df，即每一个子df中有3个交易日的交易数据
        btc_siblings = [btc_raw.iloc[sib_ind * 3:(sib_ind + 1) * 3, :]
                        for sib_ind in np.arange(0, int(btc_raw.shape[0] / 3))]

        btc_df = pd.DataFrame()
        for sib_btc in btc_siblings:
            # 使用数据标准化将连续3天交易日中的连续数值特征进行标准化操作
            sib_btc_scale = ABuScalerUtil.scaler_std(
                sib_btc.filter(['open', 'close', 'high', 'low', 'volume', 'pre_close',
                                'ma5', 'ma10', 'ma21', 'ma60', 'atr21', 'atr14']))
            # 把标准化后的和big_wave，date_week连接起来
            sib_btc_scale = pd.concat([sib_btc['big_wave'], sib_btc_scale, sib_btc['date_week']], axis=1)

            # 抽取第一天，第二天的大多数特征分别改名字以one，two为特征前缀，如：one_open，one_close，two_ma5，two_high.....
            a0 = sib_btc_scale.iloc[0].filter(['open', 'close', 'high', 'low', 'volume', 'pre_close',
                                               'ma5', 'ma10', 'ma21', 'ma60', 'atr21', 'atr14', 'date_week'])
            a0.rename(index={'open': 'one_open', 'close': 'one_close', 'high': 'one_high', 'low': 'one_low',
                             'volume': 'one_volume', 'pre_close': 'one_pre_close',
                             'ma5': 'one_ma5', 'ma10': 'one_ma10', 'ma21': 'one_ma21',
                             'ma60': 'one_ma60', 'atr21': 'one_atr21', 'atr14': 'one_atr14',
                             'date_week': 'one_date_week'}, inplace=True)

            a1 = sib_btc_scale.iloc[1].filter(['open', 'close', 'high', 'low', 'volume', 'pre_close',
                                               'ma5', 'ma10', 'ma21', 'ma60', 'atr21', 'atr14', 'date_week'])
            a1.rename(index={'open': 'two_open', 'close': 'two_close', 'high': 'two_high', 'low': 'two_low',
                             'volume': 'two_volume', 'pre_close': 'two_pre_close',
                             'ma5': 'two_ma5', 'ma10': 'two_ma10', 'ma21': 'two_ma21',
                             'ma60': 'two_ma60', 'atr21': 'two_atr21', 'atr14': 'two_atr14',
                             'date_week': 'two_date_week'}, inplace=True)
            # 第三天的特征只使用'open', 'low', 'pre_close', 'date_week'，该名前缀today，如today_open，today_date_week
            a2 = sib_btc_scale.iloc[2].filter(['big_wave', 'open', 'low', 'pre_close', 'date_week'])
            a2.rename(index={'open': 'today_open', 'low': 'today_low',
                             'pre_close': 'today_pre_close',
                             'date_week': 'today_date_week'}, inplace=True)
            # 将抽取改名字后的特征连接起来组合成为一条新数据，即3天的交易数据特征－>1条新的数据
            btc_df = btc_df.append(pd.concat([a0, a1, a2], axis=0), ignore_index=True)
        return btc_df
